import os
import kmviz.core.config as kconf
from kmviz.core import KmVizError
from kmviz.core.log import setup_logger

from kmviz.init import make_kmviz_app

config_file = os.environ["KMVIZ_CONF"]
log_directory = os.environ["KMVIZ_LOG_DIR"]
log_stderr = os.environ["KMVIZ_LOG_STDERR"]
verbosity = os.environ["KMVIZ_VERBOSITY"]

if not config_file:
    raise KmVizError("'KMVIZ_CONF' env variable not found")

if not log_directory:
    log_directory = ".kmviz_log"

if not verbosity:
    verbosity = "INFO"

if not log_stderr:
    log_stderr = True

kconf.init_global_state()

setup_logger(verbosity, log_directory=log_directory, with_stderr=log_stderr, traceback=None)

kconf.st.configure("db", config_file)
app = make_kmviz_app(kconf.st)
app = app.server









from kmviz.core.notifier.notifier import Notifier, is_email
from kmviz.core.query.query import Query
from kmviz.core import KmVizQueryError
import sendgrid
from sendgrid.helpers.mail import Mail, Email, To, Content
from kmviz.core.provider.options import TextOption, update_option
import re

class SendGridNotifier(Notifier):
    def __init__(self,  success: str, failure: str, subject: str, subject_failure: str, api_key: str, sender: str, custom = {}):
        super().__init__(success, failure, subject, subject_failure, custom)
        self._api = api_key
        self._sender = sender
        self._client = sendgrid.SendGridAPIClient(api_key=self._api)

    def send_success(self, session: str, options: dict, results: dict = {}):
        c = Content("text/plain", self.format_success(session))
        m = Mail(Email(self._sender), To(options["Email"]), self.format_subject(session), c)
        return self._client.client.mail.send.post(request_body=m.get())

    def send_failure(self, session: str, options: dict, reason: str = None):
        c = Content("text/plain", self.format_failure(session, reason))
        m = Mail(Email(self._sender), To(options["Email"]), self.format_subject_failure(session, reason), c)
        return self._client.client.mail.send.post(request_body=m.get())

    def check(self, options: dict):
        if not is_email(options["Email"]):
            raise KmVizQueryError(f"'{options['Email']}': not an email.")

    def options(self):
        return {"Email": update_option(TextOption("Email", "", "Your email"), {"is_required": True})}




from dash_extensions.enrich import html, dcc
import dash_mantine_components as dmc
import dash_ag_grid as dag
from typing import Dict, Any, Union, List, Optional
from kmviz.ui.utils import icons
import dash_bootstrap_components as dbc

import dash_daq as daq
IDType = Union[str, Dict[str, Any]]

from dash_mantine_components import __version__ as dmc_version

dmc_old = dmc_version == "0.12.1"
dmc_new = not dmc_old

def classname(base: str, idd: Optional[IDType], kwargs: Dict[str, Any]) -> str:
    if idd is None:
        sid = ""
    else:
        sid = idd if isinstance(idd, str) else f"{idd['type']}-{idd['index']}"

    c = f"{base} {sid}"
    if kwargs and "className" in kwargs:
        c += f" {kwargs['className']}"
    return c

def patch_dmc_value(old, new, d):
    if dmc_new:
        return
    if old in d:
        if not new:
            del d[old]
        else:
            tmp = d[old]
            del d[old]
            d[new] = tmp

def patch_dmc(kwargs: dict):
    patch_dmc_value("leftSection", "icon", kwargs)
    patch_dmc_value("decimalScale", "precision", kwargs)
    patch_dmc_value("gap", "spacing", kwargs)

def forward(kwargs: dict):
    if kwargs and "className" in kwargs:
        del kwargs["className"]

    if dmc_old:
        patch_dmc(kwargs)
    return kwargs

def dmc_select_version(old, new):
    if dmc_version == "0.12.1":
        return old
    return new

km_color = "#1C7ED6"
kshow = {"display": "inline"}
khide = {"display": "none"}

class ComponentFactory:

    _cls_prefix = "kmviz"
    _button = f"{_cls_prefix}-button"
    _select = f"{_cls_prefix}-select"
    _multi = f"{_cls_prefix}-multi"
    _slider = f"{_cls_prefix}-slider"
    _switch = f"{_cls_prefix}-switch"
    _number = f"{_cls_prefix}-number"
    _json = f"{_cls_prefix}-json"
    _text = f"{_cls_prefix}-text"
    _area = f"{_cls_prefix}-area"
    _color = f"{_cls_prefix}-color"

    _tabs = f"{_cls_prefix}-tabs"
    _tabs_list = f"{_cls_prefix}-tabs-list"
    _tabs_tab = f"{_cls_prefix}-tabs-tab"
    _tabs_panel = f"{_cls_prefix}-tabs-panel"

    _upload = f"{_cls_prefix}-upload"

    _group = f"{_cls_prefix}-group"

    _div = f"{_cls_prefix}-div"

    _h1 = f"{_cls_prefix}-h1"
    _h2 = f"{_cls_prefix}-h2"
    _h3 = f"{_cls_prefix}-h3"
    _h4 = f"{_cls_prefix}-h4"

    _center = f"{_cls_prefix}-center"
    _divider = f"{_cls_prefix}-divider"


    _action = f"{_cls_prefix}-action"
    _segmented = f"{_cls_prefix}-segmented"

    _ag_grid = f"{_cls_prefix}-ag-grid"

    @staticmethod
    def _make_select_data(data: Union[List[str], List[Dict[str, str]]]):
        if data and isinstance(data[0], str):
            return [{ "label": value, "value": value } for value in data]
        return data

    @classmethod
    def button(cls, id: IDType, children: Any, **kwargs) -> dmc.Button:
        patch_dmc_value("leftSection", "leftIcon", kwargs)
        return dmc.Button(
            children,
            id=id,
            className=classname(cls._button, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def action(cls, id: IDType, icon: Any, **kwargs) -> dmc.ActionIcon:
        return dmc.ActionIcon(
            icon,
            id=id,
            className=classname(cls._action, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def color(cls, id: IDType, **kwargs):
        if dmc_old:
            return cls._old_color(id, **kwargs)
        else:
            return dmc.ColorInput(
                id=id,
                className=classname(cls._color, id, kwargs),
                **forward(kwargs)
            )

    @classmethod
    def _old_color(cls, id: IDType, **kwargs):
        format = None if "noTranspa" in kwargs else "rgba"
        return html.Div([
            dmc.Badge(kwargs["label"], size="sm", id=f"{str(id)}-target", className=classname(cls._color, id, kwargs)),
            dbc.Popover(
                dmc.ColorPicker(id=id, value=kwargs["value"] if "value" in kwargs else None, format=format),
                target=f"{str(id)}-target",
                trigger="hover"
            )
        ])

    @classmethod
    def select(cls, id: IDType, data: Union[List[str], List[Dict[str, str]]], **kwargs) -> dmc.Select:
        return dmc.Select(
            id=id,
            data=cls._make_select_data(data),
            className=classname(cls._select, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def multi(cls, id: IDType, data: Union[List[str], List[Dict[str, str]]], **kwargs) -> dmc.MultiSelect:
        return dmc.MultiSelect(
            id=id,
            data=cls._make_select_data(data),
            className=classname(cls._multi, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def slider(cls, id: IDType, **kwargs) -> dmc.Slider:
        return dmc.Slider(
            id=id,
            className=classname(cls._slider, id, kwargs)
            **forward(kwargs)
        )

    @classmethod
    def switch(cls, id: IDType, **kwargs) -> dmc.Switch:
        return dmc.Switch(
            id=id,
            className=classname(cls._switch, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def number(cls, id: IDType, **kwargs) -> dmc.NumberInput:
        return dmc.NumberInput(
            id=id,
            className=classname(cls._number, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def segmented(cls, id: IDType, data: Union[List[str], List[Dict[str, str]]], **kwargs) -> dmc.SegmentedControl:
        return dmc.SegmentedControl(
            id=id,
            data=cls._make_select_data(data),
            className=classname(cls._segmented, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def json(cls, id: IDType, **kwargs) -> dmc.JsonInput:
        return dmc.JsonInput(
            id=id,
            className=classname(cls._json, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def text(cls, id: IDType, **kwargs) -> dmc.TextInput:
        return dmc.TextInput(
            id=id,
            className=classname(cls._text, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def area(cls, id: IDType, **kwargs) -> dmc.Textarea:
        return dmc.Textarea(
            id=id,
            className=classname(cls._area, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def group(cls, id: IDType, *components, **kwargs) -> dmc.Group:
        return dmc.Group(
            components,
            id=id,
            className=classname(cls._group, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def font(cls, id: IDType, **kwargs) -> dmc.Select:
        data = [
            "Arial", "Balto", "Courier New", "Droid Sans",
            "Droid Serif", "Droid Sans Mono", "Gravitas One",
            "Old Standard TT", "Open Sans", "Overpass",
            "PT Sans Narrow", "Raleway", "Times New Roman"
        ]
        return cls.select(id, data=data, value="Arial", leftSection=icons("family"), **kwargs)

    @classmethod
    def sseqscale(cls, id: IDType, **kwargs) -> dmc.Select:
        data = [
            'Brwnyl', 'Agsunset', 'Sunsetdark', 'Magenta', 'Sunset',
            'Purpor', 'Purp', 'Tealgrn', 'Teal', 'Bluyl', 'Aggrnyl',
            'Emrld', 'Darkmint', 'Blugrn', 'Mint', 'Pinkyl',
            'Peach', 'Oryel', 'Redor', 'Burgyl', 'Burg',
            'tempo', 'amp', 'speed', 'matter', 'algae', 'dense', 'deep',
            'gray', 'ice', 'solar', 'haline', 'thermal', 'turbid', 'YlOrRd',
            'YlOrBr', 'YlGnBu', 'YlGn', 'Reds', 'RdPu', 'RdBu', 'Purples',
            'PuRd', 'PuBuGn', 'PuBu', 'Oranges', 'OrRd', 'Greys', 'Greens',
            'GnBu', 'BuPu', 'BuGn', 'Blues', 'Rainbow', 'Jet', 'Hot', 'Electric',
            'Bluered', 'Blackbody', 'Turbo', 'Plasma', 'Magma', 'Inferno',
            'Cividis', 'Viridis', 'Plotly3'
        ]
        return cls.select(id, data=data, leftSection=icons("picker"), **kwargs)

    @classmethod
    def sdivscale(cls, id: IDType, **kwargs) -> dmc.Select:
        data = [
            'Portland', 'Picnic', 'Earth', 'Tropic', 'Tealrose', 'Temps', 'Geyser',
            'Fall', 'Armyrose', 'oxy', 'curl', 'delta', 'balance',
            'Spectral', 'RdYlGn', 'RdYlBu', 'RdGy', 'RdBu', 'PuOr', 'PiYG',
            'PRGn', 'BrBG'
        ]
        return cls.select(id, data=data, leftSection=icons("picker"), **kwargs)

    @classmethod
    def scycscale(cls, id: IDType, **kwargs) -> dmc.Select:
        data = ['mygbm', 'mrybm', 'HSV', 'Phase', 'Edge', 'IceFire', 'Twilight']
        return cls.select(id, data=data, leftSection=icons("picker"), **kwargs)

    @classmethod
    def div(cls, id: IDType, *children, **kwargs) -> html.Div:
        return html.Div(
            children=children,
            id=id,
            className=classname(cls._div, id, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def h1(cls, text: str, id: Optional[IDType]=None, **kwargs) -> html.H1:
        return html.H1(text, className=classname(cls._h1, id, kwargs), **forward(kwargs))

    @classmethod
    def h2(cls, text: str, id: Optional[IDType]=None, **kwargs) -> html.H2:
        return html.H2(text, className=classname(cls._h2, id, kwargs), **forward(kwargs))

    @classmethod
    def h3(cls, text: str, id: Optional[IDType]=None, **kwargs) -> html.H3:
        return html.H3(text, className=classname(cls._h3, id, kwargs), **forward(kwargs))

    @classmethod
    def h4(cls, text: str, id: Optional[IDType]=None, **kwargs) -> html.H4:
        return html.H4(text, className=classname(cls._h4, id, kwargs), **forward(kwargs))

    @classmethod
    def divider(cls, **kwargs) -> dmc.Divider:
        return dmc.Divider(
            className=classname(cls._divider, None, kwargs),
            **forward(kwargs)
        )

    @classmethod
    def upload(cls, id: IDType, content: Any) -> dcc.Upload:
        return dcc.Upload(
            content,
            id=id,
            className=classname(cls._upload, id, {})
        )

    @classmethod
    def store(cls, id: IDType, data: Optional[Any]=None, **kwargs) -> dcc.Store:
        return dcc.Store(id=id, data=data, **kwargs)

    @classmethod
    def tabs(cls, id: IDType, *content: Any, value: Optional[str], **kwargs) -> dmc.Tabs:
        return dmc.Tabs(content, value=value, id=id, className=classname(cls._tabs, id, kwargs), **forward(kwargs))

    @classmethod
    def tabs_list(cls, id: IDType, *content: Any, **kwargs) -> dmc.TabsList:
        return dmc.TabsList(content, id=id, className=classname(cls._tabs_list, id, kwargs), **forward(kwargs))

    @classmethod
    def tabs_tab(cls, id: IDType, content: Any, value: str, **kwargs):# -> dmc.TabsTab:
        if dmc_old:
            return dmc.Tab(content, value=value, id=id, className=classname(cls._tabs_tab, id, kwargs), **forward(kwargs))
        else:
            return dmc.TabsTab(content, value=value, id=id, className=classname(cls._tabs_tab, id, kwargs), **forward(kwargs))

    @classmethod
    def tabs_panel(cls, id: IDType, content: Any, value: str, **kwargs):#! -> dmc.TabsTab:
        return dmc.TabsPanel(content, value=value, id=id, className=classname(cls._tabs_panel, id, kwargs), **forward(kwargs))

    @classmethod
    def ag_grid(cls, id: IDType,
                grid_opt: Dict[str, Any]={},
                col_def: Dict[str, Any]={},
                export: Dict[str, Any]={}, **kwargs) -> dag.AgGrid:

        #c = "ag-theme-balham" + " " + classname(cls._ag_grid, id, kwargs)
        c = "ag-theme-balham"

        _default = {
            "columnSize": "autoSize",
            "className": c
        }

        _default_col_def = {
            "filter": True,
            "sortable": True,
            "floatingFilter": True
        }

        _default_grid_opt = {
            "suppressMenuHide": False,
            "animateRows": False,
            "pagination": True,
            "loadingOverlayComponent": "CustomLoadingOverlay",
            "loadingOverlayComponentParams": {
                "loadingMessage": "Waiting for results..."
            },
            "rowSelection": "single",
            "enableCellTextSelection": True,
            "ensureDomOrder": True
        }

        _default_export = {
            "fileName": "data.tsv",
            "columnSeparator": "\t"
        }

        return dag.AgGrid(
            id=id,
            defaultColDef={**_default_col_def, **col_def },
            dashGridOptions={**_default_grid_opt, **grid_opt},
            csvExportParams={**_default_export, **export},
            **{**_default, **forward(kwargs)}
        )




from typing import Tuple, List, Iterable
from executor import execute, ExternalCommandFailed
from kmviz.core.io import KmVizInvalidQuery

def covxb_from_covxk(cov: Iterable[int], k: int, size: int) -> Tuple[float, List[int]]:
    covxb = [0 for _ in range(size)]
    for i, value in enumerate(cov):
        if value:
            covxb[i:i+k] = [value] * k
    return sum(covxb) / size, covxb

def covyb_from_covyk(cov: Iterable[int], k: int, size: int) -> Tuple[float, List[int]]:
    covyb = [0 for _ in range(size)]
    n = 0
    m = 0
    for i, value in enumerate(cov):
        if value:
            for j in range(i, i+k):
                covyb[j] += value
            n += 1
            m += value

    if cov[-1]:
        n += (k-1)

    begin = [covyb[x-1] / x for x in range(1, k)]
    mid = [x / k for x in covyb[k-1:size-k+1]]
    end = [covyb[x] / (k-(i)) for i, x in enumerate(range(size-k, size-1))]
    covyb = begin + mid + end

    nk = size - k + 1
    return (m / nk), (n / size), sum(covyb) / size, covyb


def make_cmd(executable: str, subcmd: str, prefix: str="", *args, **kwargs):
    cmd = executable

    if subcmd:
        cmd += f" {subcmd}"

    for k, v in kwargs.items():
        if v is not None:
            cmd += f" {prefix}{k} {v}"
        else:
            cmd += f" {prefix}{k}"

    for e in args:
        cmd += f" {args}"

    return cmd

def exec_cmd(cmd: str, **options):
    try:
        return execute(cmd, **options)
    except ExternalCommandFailed as ec:
        raise KmVizInvalidQuery(str(ec))



from .provider import Providers, Provider
from .kmindex import KmindexServerProvider, KmindexCLIProvider

PROVIDERS = {
    "kmindex-server": KmindexServerProvider,
    "kmindex-cli": KmindexCLIProvider
}



from executor import execute, ExternalCommandFailed
from kmviz.core.io import KmVizInvalidQuery

def make_cmd(executable: str, subcmd: str, *args, **kwargs):
    cmd = executable

    if subcmd:
        cmd += f" {subcmd}"

    for k, v in kwargs.items():
        if v is not None:
            cmd += f" {k} {v}"
        else:
            cmd += f" {k}"

    for e in args:
        cmd += f" {args}"

    return cmd

def exec_cmd(cmd: str, **options):
    try:
        return execute(cmd, **options)
    except ExternalCommandFailed as ec:
        raise KmVizInvalidQuery(str(ec))


from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.id_factory import IDFactory
from kmviz.ui.utils import icons
from dash_extensions.enrich import html, callback, Input, Output, no_update
from dash import Patch
from kmviz.ui.components.helpers import from_json_list, magic_json

class ShapeLayout:
    def __init__(self, factory: IDFactory, figure_id):
        self.f = factory
        self.fid = figure_id

    def layout(self) -> html.Div:
        return cf.div(
            self.f["shape-div"],
            cf.json(
                self.f("shapes"),
                placeholder='[ { "type": "circle", "x0": 0, "x1": 2, "y0": 0.1, "y1": 0.8 } ]',
                label="Shape parameters",
                value=None,
                autosize=True,
                debounce=2,
                validationError="Invalid json",
                leftSection=icons("json")
            )
        )

    def callbacks(self) -> None:

        @callback(
            Input(self.f("shapes"), "value"),
            Output(self.fid, "figure"),
            prevent_initial_call=True
        )
        def update_shapes(shapes):
            p = Patch()
            if not shapes:
                p["layout"]["shapes"] = []
            js = from_json_list(shapes, "shapes")

            if not js:
                return no_update

            all_shapes = [magic_json(x) for x in js]
            p["layout"]["shapes"] = all_shapes
            return p




import base64
from io import StringIO, BytesIO

from typing import Union, Optional
from pathlib import Path

import pandas as pd

from kmviz.core.io import KmVizIOError

import orjson

def load_file_content(filename, contents, decode: Optional[str]=None) -> Union[str, bytes]:
    content_type, data = contents.split(",")
    content = base64.b64decode(data)

    if decode:
        return content.decode(decode)

    return content

def load_file_as_df(filename, contents, sep: Optional[str]=",") -> pd.DataFrame:
    data = load_file_content(filename, contents)

    filename = Path(filename)

    try:
        if filename.suffix in (".csv", ".tsv"):
            dio = StringIO(data.decode("utf-8"))
            return pd.read_csv(dio, sep=sep)
        elif filename.suffix in (".xlsx"):
            dio = BytesIO(data)
            return pd.read_excel(dio)
    except Exception as e:
        raise KmVizIOError(f"Error while loading '{filename}' [{str(e)}]")

    raise KmVizIOError(f"'{filename.suffix}' files are not supported")

def magic_json(d: dict):
    c = d.copy()
    for k, v in d.items():
        if "_" in k:
            sk, sv = k.split("_", 1)
            if sk not in c:
                c[sk] = {}
            if isinstance(v, dict):
                v = magic_json(v)
            c[sk][sv] = v
            c[sk] = magic_json(c[sk])
            del c[k]
    return c

def from_json(value, p=None):
    try:
        value = orjson.loads(value)
        if p:
            value = value[p]
        return value
    except:
        return None

def from_json_list(value, p=None):
    try:
        value = orjson.loads(value)
        if isinstance(value, list):
            return value
        if p:
            value = value[p]
        return value
    except:
        return None




from dash_extensions.enrich import html, Input, Output, State, callback, no_update, dcc
import dash
import dash_mantine_components as dmc
from kmviz.ui.layouts.sidebar import Sidebar
from kmviz.ui.layouts.tabs import Tabs
from kmviz.ui.layouts import Global

from kmviz.ui.id_factory import kid
from kmviz.ui.components.factory import dmc_new
import kmviz.core.config as kconf
from kmviz.core.log import kmv_info

Global(kconf.st).callbacks()
Sidebar(kconf.st).callbacks()
Tabs(kconf.st).callbacks()

@callback(
    Input("kmviz-session-interval", "n_intervals"),
    State(kid.input["session-button"], "n_clicks"),
    State("kmviz-session-store-page", "data"),
    Output(kid.input["session-button"], "n_clicks"),
    Output(kid.input("session"), "value"),
)
def load_session_page(n_i, n_c, data):
    kmv_info(f"Load session {str(data)}")
    if n_i == 1:
            return 500, data
    return no_update, no_update

def make_dashboard(session_id=None):
    G = Global(kconf.st)
    S = Sidebar(kconf.st)
    T = Tabs(kconf.st)

    if dmc_new:
        layout = dmc.AppShell(
            id="kmviz-shell",
            children=[
                G.layout(),
                dmc.AppShellNavbar(S.layout(), className="kmviz-sidebar"),
                dmc.AppShellMain(T.layout())
            ],
            navbar={
                "width": 250,
                "breakpoint": "lg",
                "collapsed": {"mobile": True},
            },
            transitionDuration="50",
            className="kmviz-main"
        )
    else:
        style={}
        if kconf.st.mode == "session":
            style = {"padding-left": "10px"}

        has_session = True
        if session_id and not kconf.st.cache.has(session_id):
            has_session = False

        layout = html.Div([
            G.layout(),
            dmc.Grid([
                dmc.Col(S.layout(), id=kid.kmviz["side-layout"], span="content", className="kmviz-sidebar-layout"),
                dmc.Col(T.layout(), id=kid.kmviz["main-layout"], span="auto", className="kmviz-main-layout", style = style),
                dcc.Interval(id="kmviz-session-interval", max_intervals=1, n_intervals=0, interval=100),
                dcc.Store(id="kmviz-session-store-page", data=str(session_id)),
                dmc.Modal(
                    title=dmc.Text(str(session_id), color="red"),
                    id="kmviz-load-modal",
                    zIndex=10000,
                    children=[dmc.Text(f"'{str(session_id)}' not found. Maybe your query is still running?")],
                    opened=not has_session,
                    withCloseButton=False,
                    shadow=True
                ),
            ])
        ])

    return layout

dash.register_page(__name__, path_template=kconf.st.instance_plugin[0] + "/<session_id>", path=kconf.st.instance_plugin[0], name="dashboard", title="dashboard")

layout = make_dashboard



from abc import ABC, abstractmethod
from typing import List, Dict
import pandas as pd

class MetaDB(ABC):
    def __init__(self, idx: str, geodata: Dict[str, str] = {}) -> None:
        self._geodata = geodata
        self._idx = idx

    @abstractmethod
    def connect(self):
        """
        Initialization, like database login, should take place here
        """
        ...

    @abstractmethod
    def query(self, keys: List[str]) -> pd.DataFrame:
        """
        Select a subset of metadata

        :param keys: A sequence of identifiers
        :returns: A dataframe with the metadata corresponding to the keys
        """
        ...

    @abstractmethod
    def df(self) -> pd.DataFrame:
        """
        Get all metadata

        :returns: A dataframe with all metadata
        """
        ...

    @abstractmethod
    def keys(self) -> List[str]:
        """
        Get all fields

        :returns: The list of all fields
        """
        ...

    @property
    def geodata(self) -> dict:
        """
        :returns: A dict with the name of geographical data fields
        
        Ex:
        {
          "latitude": "Lat",
          "longitude": "Long"
        }
        """
        return self._geodata

    @property
    def idx(self) -> str:
        """
        :returns: The name of the field corresponding to entry identifiers
        """
        return self._idx







from kmviz.core.notifier.notifier import Notifier, is_email
from kmviz.core.query.query import Query
from kmviz.core.provider.options import TextOption, update_option
from kmviz.core import KmVizQueryError

from smtplib import SMTP_SSL as SMTP
from email.mime.text import MIMEText

class SMTPNotifier(Notifier):
    def __init__(self,  success: str, failure: str, subject: str, subject_failure: str, server: str, sender: str, user: str, password: str, custom = {}):
        super().__init__(success, failure, subject, subject_failure, custom)
        self._server = server
        self._sender = sender
        self._client = SMTP(self._server)
        self._client.login(user, password)

    def send_success(self, session: str, options: dict, results: dict = {}):
        m = MIMEText(self.format_success(session), "text")
        m["Subject"] = self.format_subject(session)
        m["From"] = self._sender
        return self._client.sendmail(self._sender, options["Email"], m.as_string())

    def send_failure(self, session: str, options: dict, reason: str = None):
        m = MIMEText(self.format_failure(session, reason), "text")
        m["Subject"] = self.format_subject_failure(session, reason)
        m["From"] = self._sender
        return self._client.sendmail(self._sender, options["Email"], m.as_string())

    def check(self, options: dict):
        if not is_email(options["Email"]):
            raise KmVizQueryError(f"'{options['Email']}': not an email.")

    def options(self):
        return {"Email": update_option(TextOption("Email", "", "Your email"), {"is_required": True})}







import dash

import kmviz.core.config as kconf

def make_instance(plugin):
    if plugin:
        return plugin.instance()

_, plugin = kconf.st.instance_plugin

if plugin:
    dash.register_page(__name__, path="/", name=plugin.name(), title=plugin.name())

layout = make_instance(plugin)



from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, khide, kshow
from dash_extensions.enrich import html, dcc, callback, Input, Output, State, ctx, no_update
import plotly.graph_objects as go
import plotly.io as pio
from kmviz.ui.utils import icons, make_select_data
from dash.exceptions import PreventUpdate
from dash import Patch
from kmviz.core.log import kmv_debug

from kmviz.ui.id_factory import kid
from kmviz.ui.layouts.figure.plotly import px_options, valid_input, fix_px_params, make_plot_px, to_bool, select_cscale
from kmviz.ui.layouts.figure.title import TitleLayout
from kmviz.ui.layouts.figure.axe import AxesLayout
from kmviz.ui.layouts.figure.legend import LegendLayout
from kmviz.ui.layouts.figure.color import ColorLegendLayout
from kmviz.ui.layouts.figure.shape import ShapeLayout
from kmviz.ui.layouts.figure.download_graph import DownloadGraphLayout
from kmviz.ui.layouts.figure.slider import SliderLayout
from kmviz.ui.layouts.figure import apply_presets
from kmviz.ui.layouts.filter import FilterLayout

from kmviz.ui.components.helpers import from_json
from kmviz.ui.presets import set_preset_values
from kmviz.core.config import state

from dash_iconify import DashIconify

import pandas as pd
from pandas.api.types import is_numeric_dtype

import dash_mantine_components as dmc

class TraceLayout:

    def __init__(self, st: state, factory, figure_id, figure_data, preset):
        self.st = st
        self.f = factory
        self.fid = figure_id
        self.fdata = figure_data
        self.preset = preset

    def _data_panel(self):
        return cf.tabs_panel(
            self.f["data-panel"],
            cf.group(
                self.f["data-grp-1"],
                cf.multi(self.f("xselect"), [], label="X", clearable=True, searchable=True, value=[], className="kmviz-figure-select", size="xs"),
                cf.multi(self.f("yselect"), [], label="Y", clearable=True, searchable=True, value=[], className="kmviz-figure-select", size="xs"),
                cf.multi(self.f("zselect"), [], label="Z", clearable=True, searchable=True, value=[], className="kmviz-figure-select", size="xs"),
                cf.select(self.f("color"), [], label="Color", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("size"), [], label="Size", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("text"), [], label="Text", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("symbol"), [], label="Symbol", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("pattern_shape"), [], label="Pattern shape", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("base"), [], label="Base", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("line_dash"), [], label="Line dash", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("line_group"), [], label="Line group", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.multi(self.f("dimensions"), [], label="Dimensions", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("values"), [], label="Values", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("names"), [], label="Names", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="data"
        )

    def _anim_panel(self):
        return cf.tabs_panel(
            self.f["anim-panel"],
            cf.group(
                self.f["anim-grp-1"],
                cf.select(self.f("animation_frame"), [], label="Frame", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("animation_group"), [], label="Group", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="anim"
        )

    def _dist_panel(self):
        return cf.tabs_panel(
            self.f["dist-panel"],
            cf.group(
                self.f["dist-grp-1"],
                cf.select(self.f("violinmode"), ["group", "overlay"], label="Violin mode", value="group", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("boxmode"), ["group", "overlay"], label="Box mode", value="group", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("points"), ["outliers", "suspectedoutliers", "all", "False"], label="Points", value="outliers", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("box"), ["True", "False"], label="Violin box", value="False", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("notched"), ["True", "False"], label="Box notches", value="False",  clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="dist"
        )

    def _dens_panel(self):
        return cf.tabs_panel(
            self.f["dens-panel"],
            cf.group(
                self.f["dens-grp-1"],
                cf.select(self.f("histfunc"), ["count", "sum", "avg", "min", "max"], label="Hist function", value="count", searchable=True, clearable=True, size="xs", className="kmviz-figure-select"),
                cf.select(self.f("histnorm"), ["percent", "probability", "density", "probability density"], label="Hist norm", value="", searchable=True, clearable=True, size="xs", className="kmviz-figure-select"),
                cf.number(self.f("nbinsx"), label = "X bins", min=0, max=1000, step=1, value=4, icon=icons("integer"), size="xs"),
                cf.number(self.f("nbinsy"), label = "Y bins", min=0, max=1000, step=1, value=4, icon=icons("integer"), size="xs"),
                cf.number(self.f("nbins"), label = "N bins", min=0, max=1000, step=1, value=4, icon=icons("integer"), size="xs"),
                cf.select(self.f("contours_coloring"), ["fill", "heatmap", "lines", "none"], label="Contours coloring", searchable=True, clearable=True, size="xs", className="kmviz-figure-select"),
                cf.segmented(self.f("cumulative"), data=[{"label":"cumulative", "value":"True"}, {"label":"noncumulative", "value":"False"}], value="False", size="xs", className="kmviz-figure-segmented"),
            ),
            value="dens"
        )

    def _facet_panel(self):
        return cf.tabs_panel(
            self.f["facet-panel"],
            cf.group(
                self.f["facet-grp-1"],
                cf.select(self.f("facet_row"), [], label="Row", clearable=True, size="xs", className="kmviz-figure-select"),
                cf.select(self.f("facet_col"), [], label="Column", clearable=True, size="xs", className="kmviz-figure-select"),
                cf.number(self.f("facet_col_wrap"), label="Wrap", min=0, max=10, step=1, value=0, icon=icons("integer"), size="xs"),
                cf.number(self.f("facet_row_spacing"), label="Row spacing", min=0.0, max=1.0, step=0.01, value=0.03, icon=icons("floating"), decimalScale=2, size="xs"),
                cf.number(self.f("facet_col_spacing"), label="Col spacing", min=0.0, max=1.0, step=0.01, value=0.03, icon=icons("floating"), decimalScale=2, size="xs"),
            ),
            value="facet"
        )

    def _marg_panel(self):
        return cf.tabs_panel(
            self.f["marg-panel"],
            cf.group(
                self.f["marg-grp-1"],
                cf.select(self.f("marginal_x"), ["rug", "box", "violin", "histogram"], label="X", clearable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("marginal_y"), ["rug", "box", "violin", "histogram"], label="Y", clearable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="marg"
        )

    def _trend_panel(self):
        return cf.tabs_panel(
            self.f["trend-panel"],
            [
                cf.group(
                    self.f["trend-grp-1"],
                    cf.select(self.f("trendline"), ["ols", "lowess", "rolling", "expanding", "ewm"], label="Type", clearable=True, className="kmviz-figure-select", size="xs"),
                    cf.select(self.f("trendline_scope"), ["trace", "overall"], label="Scope", value="trace", className="kmviz-figure-select", size="xs"),
                ),
                cf.json(self.f("trendline_options"), placeholder="ex: {'value1': 'circle-open', 'value2': 'square-open'}", label="Options", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json")),
            ],
            value="trend"
        )

    def _style_panel(self):
        return cf.tabs_panel(
            self.f["style-panel"],
            [
                cf.group(
                    self.f["style-grp-1"],
                    cf.select(self.f("barmode"), ["relative", "group", "overlay"], label="Bar mode", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                    cf.select(self.f("line_shape"),  ["hv", "vh", "hvh", "vhv", "spline", "linear"], label="Line shape", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                    cf.select(self.f("orientation"), [ { "label": "vertical", "value": "v" }, { "label": "horizontal", "value": "h" } ], label="Orientation", className="kmviz-figure-select", size="xs"),
                    cf.select(self.f("markers"), [ { "label": "Show", "value": "True" }, { "label": "Hide", "value": "False" } ], label="Markers", className="kmviz-figure-select", size="xs")
                ),
                cf.group(
                    self.f["style-grp-2"],
                    cf.select(self.f("template"), list(pio.templates), label="Theme", clearable=True, searchable=True, size="xs", className="kmviz-figure-select", leftSection=icons("style", width=20)),
                    cf.sseqscale(self.f("color_seq_continuous_scale"), label="Color sequential scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.sdivscale(self.f("color_div_continuous_scale"), label="Color diverging scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.scycscale(self.f("color_cyc_continuous_scale"), label="Color cyclical scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.number(self.f("color_continuous_midpoint"), label="Color midpoint", min=None, max=None, step=None, value=None, leftSection=icons("floating"), decimalScale=2, size="xs"),
                    cf.number(self.f("opacity"), label="Opacity", min=0.0, max=1.0, step=0.01, value=0.7, leftSection=icons("floating"), decimalScale=2, size="xs"),
                    cf.number(self.f("size_max"), label="Marker max size", min=0, max=50, step=1, value=15, leftSection=icons("integer"), decimalScale=2, size="xs", style = { "width": "100px"}),
                    cf.number(self.f("size_min"), label="Marker min size", min=0, max=50, step=1, value=2, leftSection=icons("integer"), decimalScale=2, size="xs", style = { "width": "100px"}),
                ),
                cf.group(
                    self.f["style-grp-3"],
                    cf.json(self.f("color_discrete_map"), placeholder='ex: {"v1": "blue", "v2": "red"}', label="Color map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("symbol_map"), placeholder='ex: {"v1": "circle-open", "v2": "square-open"}', label="Symbol map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("color_discrete_sequence"), placeholder='ex: {"seq": ["red", "blue"] }', label="Color sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("symbol_sequence"), placeholder='ex: {"seq": ["circle-open", "square-open"] }', label="Symbol sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    grow=True,
                ),
                cf.group(
                    self.f["style-grp-4"],
                    cf.json(self.f("pattern_shape_map"), placeholder="ex: {'value1': '+', 'value2': '/'}", label="Pattern shape map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("pattern_shape_sequence"), placeholder="ex: {'seq': ['+', '/', '-'] }", label="Pattern shape sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("line_dash_map"), placeholder="ex: {'value1': 'dot', 'value2': 'dash'}", label="Line dash map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("line_dash_sequence"), placeholder="ex: {'seq': ['longdash', 'dashdot', 'longdashdot'] }", label="Line dash sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    grow=True
                ),
            ],
            value="style"
        )

    def layout(self):
        ptype = self.st.ui.plot_types
        return cf.tabs(
            self.f["tabs"],
            cf.tabs_list(
                self.f["tabslist"],
                cf.select(self.f("type"), ptype, placeholder="Type", leftSection=icons("plot"), className="kmviz-figure-select"),
                cf.tabs_tab(self.f["data-tab"], "Data", value="data"),
                cf.tabs_tab(self.f["anim-tab"], "Animation", value="anim"),
                cf.tabs_tab(self.f["dist-tab"], "Distribution", value="dist"),
                cf.tabs_tab(self.f["dens-tab"], "Density", value="dens"),
                cf.tabs_tab(self.f["facet-tab"], "Facet", value="facet"),
                cf.tabs_tab(self.f["marg-tab"], "Marginal", value="marg"),
                cf.tabs_tab(self.f["trend-tab"], "Trendline", value="trend"),
                cf.tabs_tab(self.f["style-tab"], "Style", value="style"),
            ),
            self._data_panel(),
            self._anim_panel(),
            self._dist_panel(),
            self._dens_panel(),
            self._facet_panel(),
            self._marg_panel(),
            self._trend_panel(),
            self._style_panel(),

            value=None
        )

    def apply_plot_presets(self, params: dict, ptype, X, Y, Z, priority):
        res = dict(X=X, Y=Y, Z=Z, type=ptype)

        for v in ("X", "Y", "Z", "type"):
            if v in params:
                if params[v] is not None:
                    if priority or not len(params[v]):
                        res[v] = params[v]
                del params[v]

        return res["type"], res["X"], res["Y"], res["Z"]

    def callbacks(self):

        @callback(
            Input(self.f("type"), "value"),
            State(self.f.all, "id"),
            Output(self.f.all, "style"),
            prevent_initial_call=True
        )
        def update_constraints(ptype, options):
            if ptype is None:
                return (kshow,) + (khide,) * (len(options)-1)
            out = []

            show = {"display": "inline"}
            hide = {"display": "none"}

            for opt in options:
                if opt["index"] in px_options[ptype]:
                    out.append(show)
                elif opt["index"] == "size_min" and "Scatter" in ptype:
                    out.append(show)
                else:
                    out.append(hide)
            return out

        @callback(
            Input(self.fdata, "virtualRowData"),
            Output(self.f("xselect"), "data"),
            Output(self.f("yselect"), "data"),
            Output(self.f("zselect"), "data"),
            Output(self.f("color"), "data"),
            Output(self.f("text"), "data"),
            Output(self.f("symbol"), "data"),
            Output(self.f("animation_frame"), "data"),
            Output(self.f("animation_group"), "data"),
            Output(self.f("facet_row"), "data"),
            Output(self.f("facet_col"), "data"),
            Output(self.f("base"), "data"),
            Output(self.f("pattern_shape"), "data"),
            Output(self.f("line_dash"), "data"),
            Output(self.f("line_group"), "data"),
            Output(self.f("dimensions"), "data"),
            Output(self.f("values"), "data"),
            Output(self.f("names"), "data"),
            Output(self.f("size"), "data"),
            prevent_initial_call=True
        )
        def update_data(data):
            if not data:
                raise PreventUpdate

            df = pd.DataFrame.from_dict(data)
            cols = make_select_data(list(df))
            cols_size = []
            for c in list(df):
                if is_numeric_dtype(df[c]):
                    if all(x >= 0 for x in df[c]):
                        cols_size.append(c)

            return (cols,) * 17 + (cols_size,)

        def _blank() -> go.Figure:
            fig = go.Figure(go.Scatter(x=[], y = []))
            fig.update_layout(template = "seaborn")
            return fig

        @callback(
            Output(self.fid, "figure"),
            Input(self.fdata, "virtualRowData"),
            Input(self.f("type"), "value"),
            Input(self.f("xselect"), "value"),
            Input(self.f("yselect"), "value"),
            Input(self.f("zselect"), "value"),
            Input(self.f("size"), "value"),
            Input(self.f("color"), "value"),
            Input(self.f("text"), "value"),
            Input(self.f("symbol"), "value"),
            Input(self.f("animation_frame"), "value"),
            Input(self.f("animation_group"), "value"),
            Input(self.f("color_seq_continuous_scale"), "value"),
            Input(self.f("color_div_continuous_scale"), "value"),
            Input(self.f("color_cyc_continuous_scale"), "value"),
            Input(self.f("color_continuous_midpoint"), "value"),
            Input(self.f("symbol_map"), "value"),
            Input(self.f("symbol_sequence"), "value"),
            Input(self.f("color_discrete_map"), "value"),
            Input(self.f("color_discrete_sequence"), "value"),
            Input(self.f("opacity"), "value"),
            Input(self.f("size_max"), "value"),
            Input(self.f("size_min"), "value"),
            Input(self.f("facet_row"), "value"),
            Input(self.f("facet_col"), "value"),
            Input(self.f("facet_col_wrap"), "value"),
            Input(self.f("facet_row_spacing"), "value"),
            Input(self.f("facet_col_spacing"), "value"),
            Input(self.f("template"), "value"),
            Input(self.f("trendline"), "value"),
            Input(self.f("trendline_scope"), "value"),
            Input(self.f("trendline_options"), "value"),
            Input(self.f("marginal_x"), "value"),
            Input(self.f("marginal_y"), "value"),
            Input(self.f("base"), "value"),
            Input(self.f("pattern_shape"), "value"),
            Input(self.f("orientation"), "value"),
            Input(self.f("pattern_shape_map"), "value"),
            Input(self.f("pattern_shape_sequence"), "value"),
            Input(self.f("barmode"), "value"),
            Input(self.f("line_dash"), "value"),
            Input(self.f("line_group"), "value"),
            Input(self.f("line_shape"), "value"),
            Input(self.f("line_dash_sequence"), "value"),
            Input(self.f("line_dash_map"), "value"),
            Input(self.f("markers"), "value"),
            Input(self.f("dimensions"), "value"),
            Input(self.f("values"), "value"),
            Input(self.f("names"), "value"),
            Input(self.f("histnorm"), "value"),
            Input(self.f("histfunc"), "value"),
            Input(self.f("nbinsx"), "value"),
            Input(self.f("nbinsy"), "value"),
            Input(self.f("boxmode"), "value"),
            Input(self.f("violinmode"), "value"),
            Input(self.f("points"), "value"),
            Input(self.f("notched"), "value"),
            Input(self.f("box"), "value"),
            Input(self.f("contours_coloring"), "value"),
            Input(self.f("cumulative"), "value"),
            Input(self.f("nbins"), "value"),
            State(kid.store["session-id"], "data"),
            State(kid.kmviz("database"), "value"),
            Input(self.preset, "value"),
            prevent_initial_call=True
        )
        def make_plot(data, ptype, X, Y, Z,
                      size, color, text, symbol, animation_frame, animation_group,
                      cscs, cdcs, cccs, ccm, sm, ss, cdm, cds, opacity, size_max, size_min,
                      frow, fcol, fcolw, frows, fcols, template,
                      trendline, trendline_scope, trendline_opt, mx, my,
                      base, pattern_shape, orientation, ps_map, ps_seq, barmode,
                      line_dash, line_group, line_shape, ld_seq, ld_map, markers,
                      dimensions, values, names, histn, histf, binx, biny,
                      boxmode, violinmode, points, notched, box, contours_coloring,
                      cumulative, nbins,
                      session, database, preset):

            if not data:
                raise PreventUpdate

            kmv_debug(f"{session}: 'update_plot' triggered by '{ctx.triggered_id}'")

            df = pd.DataFrame.from_dict(data)

            markers = to_bool(markers)
            notched = to_bool(notched)
            box = to_bool(box)
            points = to_bool(points)
            cumulative = to_bool(cumulative)

            scale = select_cscale([cscs, cdcs, cccs])

            params = {
                "size": size,
                "color": color,
                "text": text,
                "symbol": symbol,
                "base": base,
                "pattern_shape": pattern_shape,
                "pattern_shape_sequence": from_json(ps_seq, "seq"),
                "pattern_shape_map": from_json(ps_map),
                "orientation": orientation,
                "barmode": barmode,
                "animation_frame": animation_frame,
                "animation_group": animation_group,
                "color_continuous_midpoint": ccm,
                "color_continuous_scale": scale,
                "symbol_map": from_json(sm),
                "symbol_sequence": from_json(ss, "seq"),
                "color_discrete_map": from_json(cdm),
                "color_discrete_sequence": from_json(cds, "seq"),
                "opacity": opacity,
                "size_max": size_max,
                "facet_row": frow,
                "facet_col": fcol,
                "facet_col_wrap": fcolw,
                "facet_row_spacing": frows,
                "facet_col_spacing": fcols,
                "template": template,
                "marginal_x": mx,
                "marginal_y": my,
                "trendline": trendline,
                "trendline_scope": trendline_scope,
                "trendline_options": from_json(trendline_opt),
                "line_dash": line_dash,
                "line_group": line_group,
                "line_shape": line_shape,
                "line_dash_sequence": from_json(ld_seq, "seq"),
                "line_dash_map": from_json(ld_map),
                "markers": markers,
                "dimensions": dimensions,
                "values": values,
                "names": names,
                "histnorm": histn,
                "histfunc": histf,
                "nbinsx": binx,
                "nbinsy": biny,
                "boxmode": boxmode,
                "violinmode": violinmode,
                "points": points,
                "notched": notched,
                "box": box,
                "contours_coloring": contours_coloring,
                "nbins": nbins,
                "cumulative": cumulative
            }

            if self.st.conf.preset == "fixed":
                if preset and preset in self.st.engine.get(database).presets.plot:
                    presets = self.st.engine.get(database).presets.plot

                    presets_params = presets[preset].model_dump()

                    if "min_size" in presets_params:
                        p_min_size = presets_params["min_size"]
                        if self.st.engine.get(database).presets.priority:
                            size_min = p_min_size
                        else:
                            if p_min_size > size_min:
                                size_min = p_min_size
                        del presets_params["min_size"]

                    ptype, X, Y, Z, = self.apply_plot_presets(presets_params, ptype, X, Y, Z, self.st.engine.get(database).presets.priority)
                    params = apply_presets(presets_params, params, self.st.engine.get(database).presets.priority)

            if not preset and not ptype:
                return _blank()

            valid_input(ptype, X, Y, Z)
            params = fix_px_params(params, ptype)
            fig = make_plot_px(ptype, df, X, Y, Z, params)

            if ("Scatter" in ptype):
                fig.update_traces(marker=dict(sizemin=size_min))

            return fig

        @callback(
            Input(self.fid, "selectedData"),
            Output(self.fdata, "filterModel"),
            prevent_initial_call=True
        )
        def on_selected(sdata):
            trigger = ctx.triggered_id

            if trigger == self.fid:
                conditions = []

                if sdata is None:
                    return {}

                if not len(sdata["points"]):
                    raise PreventUpdate

                for data in sdata["points"]:
                    conditions.append(
                        {"filter": data["hovertext"], "filterType":"text", "type":"equals"}
                    )

                p = Patch()
                p["ID"] = { "filterType": "text", "operator":"OR", "conditions": conditions }
                return p

            raise PreventUpdate


        if self.st.mode != "plot" and self.st.ui.with_sequence_tab:
            @callback(
                Input(self.fid, "clickData"),
                Output(kid.sequence["sample"], "value"),
                Output(self.fdata, "filterModel"),
                Output(kid.tabs["tabs"], "value"),
                prevent_initial_call=True,
            )
            def on_click_data_plot(value):
                sample = value["points"][0]["hovertext"]

                p = Patch()
                p["ID"] = {'filterType': 'text', 'type': 'equals', 'filter': sample}

                return sample, p, "sequence"

class PlotLayout:
    def __init__(self, st: state, factory, figure_id, figure_data):
        self.st = st
        self.f = factory
        self.fid = figure_id
        self.fdata = figure_data

        self._trace = TraceLayout(self.st, self.f.new("trace"), self.fid, self.fdata, self.f["preset"])
        self._title = TitleLayout(self.f.new("title"), self.fid)
        self._axes = AxesLayout(self.f.new("axes"), self.fid)
        self._slider = SliderLayout(self.f.new("slider"), self.fid, self._axes.x.f("axis-index"))
        self._legend = LegendLayout(self.f.new("legend"), self.fid)
        self._clegend = ColorLegendLayout(self.f.new("clegend"), self.fid)
        self._shape = ShapeLayout(self.f.new("shape"), self.fid)
        self._dl = DownloadGraphLayout(self.f.new("download"), self.fid, "plot")
        self._filter = FilterLayout(self.f.new("filter"), self.fdata, self.st.ui.plot_filter_message)

    def _blank(self) -> go.Figure:
        fig = go.Figure(go.Scatter(x=[], y = []))
        fig.update_layout(template = "seaborn")
        return fig

    def _make_graph(self) -> dcc.Graph:
        return dcc.Loading(id=self.f("loading"), type="default", delay_show=250, children=[
            dcc.Graph(
                id=self.fid,
                figure=self._blank(),
                responsive=True,
                mathjax=True,
                className="kmviz-dcc-plot-graph",
                config = {
                    "modeBarButtonsToAdd": [
                        "drawline", "drawopenpath", "drawclosedpath", "drawcircle", "drawrect", "eraseshape"
                    ]
                }
            )
        ])

    def _make_tabs(self):
        return cf.tabs(
            self.f["tabs"],
            cf.tabs_list(
                self.f["tabslist"],
                cf.tabs_tab(self.f["trace-tab"], "Trace", value="trace"),
                cf.tabs_tab(self.f["title-tab"], "Title", value="title"),
                cf.tabs_tab(self.f["axes-tab"], "Axes", value="axes"),
                cf.tabs_tab(self.f["slider-tab"], "Slider", value="slider"),
                cf.tabs_tab(self.f["legend-tab"], "Legend", value="legend"),
                cf.tabs_tab(self.f["clegend-tab"], "Colorbar", value="clegend"),
                cf.tabs_tab(self.f["shape-tab"], "Shape", value="shape"),
                self._dl.layout(),
                dmc.Space(w=10),
                self._filter.layout(),
                cf.select(self.f["preset"], [], placeholder="Preset", clearable=True, icon=icons("preset"), size="xs", className="kmviz-figure-select kmviz-preset-select", style=khide),
            ),
            cf.tabs_panel(self.f["trace-panel"], self._trace.layout(), value="trace"),
            cf.tabs_panel(self.f["title-panel"], self._title.layout(), value="title"),
            cf.tabs_panel(self.f["axes-panel"], self._axes.layout(), value="axes"),
            cf.tabs_panel(self.f["slider-panel"], self._slider.layout(), value="slider"),
            cf.tabs_panel(self.f["legend-panel"], self._legend.layout(), value="legend"),
            cf.tabs_panel(self.f["clegend-panel"], self._clegend.layout(), value="clegend"),
            cf.tabs_panel(self.f["shape-panel"], self._shape.layout(), value="shape"),
            value="trace"
        )

    def layout(self) -> html.Div:
        return cf.div(
            self.f["div"],
            self._make_graph(),
            self._make_tabs(),
        )

    def callbacks(self) -> None:
        self._trace.callbacks()
        self._title.callbacks()
        self._axes.callbacks()
        self._legend.callbacks()
        self._shape.callbacks()
        self._dl.callbacks()
        self._slider.callbacks()
        self._clegend.callbacks()
        self._filter.callbacks()

        @callback(
            Input(self.f["rmf"], "n_clicks"),
            Output(self.fdata, "filterModel")
        )
        def rmf(n_clicks):
            if n_clicks:
                return {}
            raise PreventUpdate


        @callback(
            Input(kid.kmviz("database"), "value"),
            Output(self.f["preset"], "data"),
            Output(self.f["preset"], "style"),
            Output(self.f["preset"], "value"),
            prevent_initial_call=True
        )
        def update_presets(db):
            if db.startswith("__kmviz_df") or self.st.mode != "db":
                raise PreventUpdate

            presets = self.st.engine.get(db).presets

            if presets.plot:
                return make_select_data(list(presets.plot.keys())), kshow, presets.defaults.plot
            else:
                return [], khide, None


        if self.st.conf.preset == "flex":
            @callback(
                Input(self.f["preset"], "value"),
                State(kid.kmviz("database"), "value"),
                State(self._trace.f.all, "id"),
                Output(self._trace.f.all, "value"),
                State(self._title.f.all, "id"),
                Output(self._title.f.all, "value"),
                State(self._legend.f.all, "id"),
                Output(self._legend.f.all, "value"),
                prevent_initial_call=True
            )
            def update_preset_values(preset, db, ids_plot, ids_title, ids_legend):
                if db and preset in self.st.engine.get(db).presets.plot:
                    p = self.st.engine.get(db).presets.plot[preset]
                    rplot = set_preset_values(ids_plot, p.model_dump(exclude=["title", "legend"]))
                    rtitle = set_preset_values(ids_title, p.title.model_dump())
                    rlegend = set_preset_values(ids_legend, p.legend.model_dump())
                    return rplot, rtitle, rlegend
                return (no_update,) * len(ids_plot), (no_update,) * len(ids_title), (no_update,) * len(ids_legend)



import uplink
import json
import pandas as pd
import orjson
import tempfile
import shutil
from typing import Optional, Union, List
from .provider import Provider
from .options import RangeOption
from kmviz.core.query import Query, Response, QueryResponse
from kmviz.core.utils import covxb_from_covxk, covyb_from_covyk, make_cmd, exec_cmd
from kmviz.core.io import rm

class KmindexProvider(Provider):
    def __init__(self, name: str):
        super().__init__(name)

        self.options = {
            "z": RangeOption("z", 0, min=0, max=5, step=1),
            "coverage": RangeOption("coverage", 0.1, min=0.0, max=1.0, step=0.05),
        }

    def has_abs(self) -> bool:
        return self.index_infos[list(self.index_infos.keys())[0]]["bw"] > 1

    def kmer_size(self) -> int:
        smer_size = self.index_infos[list(self.index_infos.keys())[0]]["smer_size"]
        return smer_size

    def samples(self, index=None):
        if index:
            return self.index_infos[index]
        else:
            res = {}
            for k, v in self.index_infos.items():
                res[k] = v["Samples"]
            return res

    def index_list(self) -> List[str]:
        return list(self.index_infos.keys())

    @property
    def infos_df(self) -> pd.DataFrame:
        key = list(self.infos().keys())[0]
        cols = ["index"] + list(self.infos()[key].keys())
        cols.remove("samples")

        df = pd.DataFrame(columns=cols)
        for k, v in self.infos().items():
            v["index"] = k
            df.loc[len(df)] = v

        return df

    def _response_from_dict(self, response: dict, size: int, z: int) -> Response:
        if self.has_abs():
            return self._response_from_dict_abs(response, size, z)
        else:
            return self._response_from_dict_pa(response, size, z)

    def _response_from_dict_pa(self, response: dict, size: int, z) -> Response:
        xk = response["R"]
        covxk = response["P"]

        xb, covxb = covxb_from_covxk(covxk, self.kmer_size() + z, size)

        return Response(
            self.kmer_size() + z,
            xk,
            None,
            covxk,
            None,
            xb,
            None,
            covxb,
            None
        )

    def _response_from_dict_abs(self, response: dict, size: int, z: int) -> Response:
        covyk = response["P"]
        xk = response["R"]

        yk, xb, yb, covyb = covyb_from_covyk(covyk, self.kmer_size() + z, size)

        return Response(
            self.kmer_size() + z,
            xk,
            yk,
            None,
            covyk,
            xb,
            yb,
            None,
            covyb,
        )

    def _from_json(self, query: Query, options, rj) -> QueryResponse:
        responses = {}
        for k, v in rj.items():
            for qid, results in v.items():
                for sample, d in results.items():
                    responses[sample] = self._response_from_dict(d, len(query.seq), options["z"].value)

        metadata = self.db.query(responses.keys())

        covxks = []
        covxbs = []
        covyks = []
        covybs = []

        for r in metadata["ID"]:
            covxks.append(round(responses[r].xk, 3))
            covxbs.append(round(responses[r].xb, 3))

            if self.has_abs():
                covyks.append(round(responses[r].yk, 3))
                covybs.append(round(responses[r].yb, 3))

        metadata.insert(1, "CovXK", covxks, True)
        metadata.insert(2, "CovXB", covxbs, True)

        if self.has_abs():
            metadata.insert(3, "CovYK", covyks , True)
            metadata.insert(4, "CovYB", covybs, True)

        return QueryResponse(
            query,
            responses,
            metadata
        )

class KmindexCLIProvider(KmindexProvider):
    def __init__(self, name: str, index_path: str, executable: Optional[str]=None):
        super().__init__(name)
        self._bin = executable if executable else "kmindex"
        self._index_path = index_path
        self._tmp = tempfile.TemporaryDirectory()

    def _make_kmindex_infos_cmd(self):
        return make_cmd(self._bin, "index-infos", "--", index=self._index_path)

    def _make_kmindex_query_cmd(self, **arguments):
        return make_cmd(self._bin, "query", "--", index=self._index_path, verbose="error", **arguments)

    def _execute(self, cmd: str, **options):
        return exec_cmd(cmd, **options)

    def _get_tmp_paths(self, idx: str, query: Query):
        p = f"{self._tmp.name}/{idx}-{query.name}"
        return p, p + ".fa"

    def _make_options(self, fastx, output, options):
        kwargs = {}
        kwargs["fastx"] = fastx
        kwargs["output"] = output
        kwargs["zvalue"] = options["z"].value
        kwargs["threshold"] = options["coverage"].value
        kwargs["format"] = "json_vec"
        kwargs["aggregate"] = None
        kwargs["fast"] = None
        return kwargs

    def _make_kmindex_query(self, query: Query, options, idx) -> str:
        output, fastx = self._get_tmp_paths(idx, query)

        with open(fastx, "w") as stream:
            stream.write(f">{query.name}\n")
            stream.write(f"{query.seq}\n")

        kwargs = self._make_options(fastx, output, options)
        cmd = self._make_kmindex_query_cmd(**kwargs)
        self._execute(cmd)

        response = {}
        for i in self.index_list():
            with open(f"{output}/{i}.json") as jin:
                response[i] = orjson.loads(jin.read())[i]

        res = self._from_json(query, options, response)

        rm(fastx, output)

        return res

    def connect(self):
        self.index_infos = orjson.loads(self._execute(self._make_kmindex_infos_cmd(), capture=True))["index"]

    def query(self, query: Query, options: dict, idx: str) -> QueryResponse:
        return self._make_kmindex_query(query, options, idx)

class KmindexServerProvider(KmindexProvider):

    class KMIndexAPI(uplink.Consumer):
        @uplink.get("kmindex/infos")
        def infos(self):
            "Fetches indexes information."

        @uplink.json
        @uplink.post("kmindex/query")
        def query(self, data: uplink.Body):
            "Query indexes."

    def __init__(self, name: str, url: str, port: int) -> None:
        super().__init__(name)
        self._url = url
        self._port = port
        self._api = self.KMIndexAPI(base_url=f"http://{self._url}:{self._port}")

    def connect(self):
        self.index_infos = dict(self._api.infos().json())["index"]

    def query(self, query: Query, options: dict, idx: str) -> QueryResponse:

        if len(query.seq) < self.kmer_size():
            return f"min sequence size is {self.kmer_size()}"

        response = self._api.query(self.make_payload(query, options))

        return self._from_json(query, options, response.json())

    def make_payload(self, query: Query, options) -> json:
        wrap = lambda x: [x] if isinstance(x, str) else x
        return {
            "id": query.name,
            "seq": wrap(query.seq),
            "index": self.index_list(),
            "r": options["coverage"].value,
            "z": options["z"].value,
            "format": "json_vec"
        }








from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, khide, kshow
from dash_extensions.enrich import html, dcc, callback, Input, Output, State, ctx, no_update
import plotly.graph_objects as go
import plotly.io as pio
from kmviz.ui.utils import icons, make_select_data
from dash.exceptions import PreventUpdate

from kmviz.core.log import kmv_debug
import plotly.express as px
from dash import Patch
from kmviz.ui.id_factory import kid
from kmviz.ui.layouts.figure.plotly import select_cscale
from kmviz.ui.layouts.figure.title import TitleLayout
from kmviz.ui.layouts.figure.axe import AxesLayout
from kmviz.ui.layouts.figure.legend import LegendLayout
from kmviz.ui.layouts.figure.shape import ShapeLayout
from kmviz.ui.layouts.figure.download_graph import DownloadGraphLayout
from kmviz.ui.layouts.filter import FilterLayout
from kmviz.ui.layouts.figure import apply_presets, apply_legend_and_title_presets
from kmviz.ui.layouts.figure.color import ColorLegendLayout
from kmviz.ui.presets import set_preset_values

from kmviz.ui.components.helpers import from_json
from kmviz.core.config import state

from dash_iconify import DashIconify

import pandas as pd
from pandas.api.types import is_numeric_dtype

import dash_mantine_components as dmc

class MapTraceLayout:

    def __init__(self, st: state, factory, figure_id, figure_data, preset):
        self.st = st
        self.f = factory
        self.fid = figure_id
        self.fdata = figure_data
        self.preset = preset

    def _data_panel(self):
        return cf.tabs_panel(
            self.f["data-panel"],
            cf.group(
                self.f["data-grp-1"],
                cf.select(self.f("color"), [], label="Color", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("size"), [], label="Size", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("text"), [], label="Text", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("symbol"), [], label="Symbol", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="data"
        )

    def _anim_panel(self):
        return cf.tabs_panel(
            self.f["anim-panel"],
            cf.group(
                self.f["anim-grp-1"],
                cf.select(self.f("animation_frame"), [], label="Frame", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
                cf.select(self.f("animation_group"), [], label="Group", clearable=True, searchable=True, className="kmviz-figure-select", size="xs"),
            ),
            value="anim"
        )

    def _style_panel(self):
        return cf.tabs_panel(
            self.f["style-panel"],
            [
                cf.group(
                    self.f["style-grp-1"],
                    cf.select(self.f("template"), list(pio.templates), label="Theme", clearable=True, searchable=True, size="xs", className="kmviz-figure-select", leftSection=icons("style", width=20)),
                    cf.select(self.f("projection"), [ 'equirectangular', 'mercator', 'orthographic', 'natural earth', 'kavrayskiy7', 'miller', 'robinson', 'eckert4', 'azimuthal equal area', 'azimuthal equidistant', 'conic equal area', 'conic conformal', 'conic equidistant', 'gnomonic', 'stereographic', 'mollweide', 'hammer', 'transverse mercator', 'albers usa', 'winkel tripel', 'aitoff', 'sinusoidal' ], label="Projection", value="equirectangular", clearable=True, searchable=True, className="kmviz-figure-select", size="xs", icon=icons("map") ),
                    cf.sseqscale(self.f("color_seq_continuous_scale"), label="Color sequential scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.sdivscale(self.f("color_div_continuous_scale"), label="Color diverging scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.scycscale(self.f("color_cyc_continuous_scale"), label="Color cyclical scale", clearable=True, searchable=True, size="xs", className="kmviz-figure-select"),
                    cf.number(self.f("color_continuous_midpoint"), label="Color midpoint", min=None, max=None, step=None, value=None, leftSection=icons("floating"), decimalScale=2, size="xs"),
                    cf.number(self.f("opacity"), label="Opacity", min=0.0, max=1.0, step=0.01, value=0.7, leftSection=icons("floating"), decimalScale=2, size="xs"),
                    cf.number(self.f("size_max"), label="Marker max size", min=0, max=50, step=1, value=15, leftSection=icons("integer"), decimalScale=2, size="xs", style = { "width": "100px"}),
                    cf.number(self.f("size_min"), label="Marker min size", min=0, max=50, step=1, value=2, leftSection=icons("integer"), decimalScale=2, size="xs", style = { "width": "100px"}),
                ),
                cf.group(
                    self.f["style-grp-2"],
                    cf.json(self.f("color_discrete_map"), placeholder='ex: {"v1": "blue", "v2": "red"}', label="Color map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("symbol_map"), placeholder='ex: {"v1": "circle-open", "v2": "square-open"}', label="Symbol map", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("color_discrete_sequence"), placeholder='ex: {"seq": ["red", "blue"] }', label="Color sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    cf.json(self.f("symbol_sequence"), placeholder='ex: {"seq": ["circle-open", "square-open"] }', label="Symbol sequence", value=None, formatOnBlur=True, maxRows=4, autosize=True, debounce=1, validationError="Invalid json", icon=icons("json"), size="xs"),
                    grow=True,
                ),
            ],
            value="style"
        )

    def layout(self):
        return cf.tabs(
            self.f["tabs"],
            cf.tabs_list(
                self.f["tabslist"],
                cf.tabs_tab(self.f["data-tab"], "Data", value="data"),
                cf.tabs_tab(self.f["anim-tab"], "Animation", value="anim"),
                cf.tabs_tab(self.f["style-tab"], "Style", value="style"),
            ),
            self._data_panel(),
            self._anim_panel(),
            self._style_panel(),

            value="data"
        )

    def callbacks(self):

        @callback(
            Input(self.fdata, "virtualRowData"),
            Output(self.f("color"), "data"),
            Output(self.f("text"), "data"),
            Output(self.f("symbol"), "data"),
            Output(self.f("animation_frame"), "data"),
            Output(self.f("animation_group"), "data"),
            Output(self.f("size"), "data"),
            prevent_initial_call=True
        )
        def update_data(data):
            if not data:
                raise PreventUpdate

            df = pd.DataFrame.from_dict(data)
            cols = make_select_data(list(df))
            cols_size = []
            for c in list(df):
                if is_numeric_dtype(df[c]):
                    if all(x >= 0 for x in df[c]):
                        cols_size.append(c)

            return (cols,) * 5 + (cols_size,)


        @callback(
            Output(self.fid, "figure"),
            Input(self.fdata, "virtualRowData"),
            Input(self.f("template"), "value"),
            Input(self.f("color"), "value"),
            Input(self.f("size"), "value"),
            Input(self.f("text"), "value"),
            Input(self.f("symbol"), "value"),
            Input(self.f("animation_frame"), "value"),
            Input(self.f("animation_group"), "value"),
            Input(self.f("projection"), "value"),
            Input(self.f("color_seq_continuous_scale"), "value"),
            Input(self.f("color_div_continuous_scale"), "value"),
            Input(self.f("color_cyc_continuous_scale"), "value"),
            Input(self.f("color_continuous_midpoint"), "value"),
            Input(self.f("symbol_map"), "value"),
            Input(self.f("symbol_sequence"), "value"),
            Input(self.f("color_discrete_map"), "value"),
            Input(self.f("color_discrete_sequence"), "value"),
            Input(self.f("opacity"), "value"),
            Input(self.f("size_max"), "value"),
            Input(self.f("size_min"), "value"),
            State(kid.store["session-id"], "data"),
            State(kid.kmviz("database"), "value"),
            State(kid.pom["store"], "data"),
            Input(self.preset, "value"),
            prevent_initial_call=True
        )
        def make_plot(data, template, color, size, text, symbol, aframe, agroup,
                      projection, cscs, cdcs, cccs, ccm, smap, sseq, cdm, cds, opacity,
                      size_max, size_min, session, database, pom, preset):

            kmv_debug(f"{session}: 'update_map' triggered by '{ctx.triggered_id}'")

            if not data:
                raise PreventUpdate

            if pom and "geodata" not in pom and not "session" in pom:
                return None

            df = pd.DataFrame.from_dict(data)

            scale = select_cscale([cscs, cdcs, cccs])

            if "geodata" in pom:
                geo = pom["geodata"]
            elif "session" in pom:
                geo = pom["session"][database]
            else:
                geo = self.st.engine.get(database).db.geodata

            if not geo:
                return None

            params = {
                "lat": geo["latitude"],
                "lon": geo["longitude"],
                "size": size,
                "color": color,
                "text": text,
                "symbol": symbol,
                "animation_frame": aframe,
                "animation_group": agroup,
                "color_continuous_midpoint": ccm,
                "color_continuous_scale": scale,
                "symbol_map": from_json(smap),
                "symbol_sequence": from_json(sseq, "seq"),
                "color_discrete_map": from_json(cdm),
                "color_discrete_sequence": from_json(cds, "seq"),
                "opacity": opacity,
                "size_max": size_max,
                "template": template,
                "projection": projection,
            }

            if self.st.conf.preset == "fixed":
                if preset and preset in self.st.engine.get(database).presets.map:
                    presets = self.st.engine.get(database).presets.map

                    md = presets[preset].model_dump(exclude=["title", "legend"])

                    if "min_size" in md:
                        p_min_size = md["min_size"]
                        if self.st.engine.get(database).presets.priority:
                            size_min = p_min_size
                        else:
                            if p_min_size > size_min:
                                size_min = p_min_size
                        del md["min_size"]

                    params = apply_presets(md, params, self.st.engine.get(database).presets.priority)
                    fig = px.scatter_geo(df, hover_name="ID", **params)
                    apply_legend_and_title_presets(fig, presets[preset])
                    return fig

            fig = px.scatter_geo(df, hover_name="ID", **params)
            fig.update_traces(marker=dict(sizemin=size_min))
            return fig

        @callback(
            Input(self.fid, "selectedData"),
            Output(self.fdata, "filterModel"),
            prevent_initial_call=True
        )
        def on_selected(sdata):
            trigger = ctx.triggered_id

            if trigger == self.fid:
                conditions = []

                if sdata is None:
                    return {}

                for data in sdata["points"]:
                    conditions.append(
                        {"filter": data["hovertext"], "filterType":"text", "type":"equals"}
                    )

                p = Patch()
                p["ID"] = { "filterType": "text", "operator":"OR", "conditions": conditions }
                return p

            raise PreventUpdate

        if self.st.mode != "plot" and self.st.ui.with_sequence_tab:
            @callback(
                Input(self.fid, "clickData"),
                Output(kid.sequence["sample"], "value"),
                Output(self.fdata, "filterModel"),
                Output(kid.tabs["tabs"], "value"),
                prevent_initial_call=True,
            )
            def on_click_data_map(value):
                sample = value["points"][0]["hovertext"]

                p = Patch()
                p["ID"] = {'filterType': 'text', 'type': 'equals', 'filter': sample}

                return sample, p, "sequence"

class MapLayout:
    def __init__(self, st: state, factory, figure_id, figure_data):
        self.st = st
        self.f = factory
        self.fid = figure_id
        self.fdata = figure_data

        self._trace = MapTraceLayout(self.st, self.f.new("trace"), self.fid, self.fdata, self.f["preset"])
        self._title = TitleLayout(self.f.new("title"), self.fid)
        self._axes = AxesLayout(self.f.new("axes"), self.fid)
        self._legend = LegendLayout(self.f.new("legend"), self.fid)
        self._clegend = ColorLegendLayout(self.f.new("clegend"), self.fid)
        self._shape = ShapeLayout(self.f.new("shape"), self.fid)
        self._dl = DownloadGraphLayout(self.f.new("download"), self.fid, "map")
        self._filter = FilterLayout(self.f.new("filter"), self.fdata, self.st.ui.map_filter_message)

    def _blank(self) -> go.Figure:
        fig = go.Figure(go.Scattergeo())
        fig.update_layout(template = "seaborn")
        fig.update_xaxes(showgrid = False, showticklabels = False, zeroline=False)
        fig.update_yaxes(showgrid = False, showticklabels = False, zeroline=False)
        return fig

    def _make_graph(self) -> dcc.Graph:
        return dcc.Loading(id=self.f("loading"), type="default", delay_show=250, children=[
            dcc.Graph(
                id=self.fid,
                figure=self._blank(),
                responsive=True,
                mathjax=True,
                className="kmviz-dcc-plot-graph",
                config = {
                    "modeBarButtonsToAdd": [
                        "drawline", "drawopenpath", "drawclosedpath", "drawcircle", "drawrect", "eraseshape"
                    ]
                }
            )
        ])

    def _make_tabs(self):
        return cf.tabs(
            self.f["tabs"],
            cf.tabs_list(
                self.f["tabslist"],
                cf.tabs_tab(self.f["trace-tab"], "Trace", value="trace"),
                cf.tabs_tab(self.f["title-tab"], "Title", value="title"),
                cf.tabs_tab(self.f["axes-tab"], "Axes", value="axes"),
                cf.tabs_tab(self.f["legend-tab"], "Legend", value="legend"),
                cf.tabs_tab(self.f["clegend-tab"], "Colorbar", value="clegend"),
                cf.tabs_tab(self.f["shape-tab"], "Shape", value="shape"),
                self._dl.layout(),
                dmc.Space(w=10),
                self._filter.layout(),
                cf.select(self.f["preset"], [], placeholder="Preset", clearable=True, icon=icons("preset"), size="xs", className="kmviz-figure-select kmviz-preset-select", style=khide),
            ),
            cf.tabs_panel(self.f["trace-panel"], self._trace.layout(), value="trace"),
            cf.tabs_panel(self.f["title-panel"], self._title.layout(), value="title"),
            cf.tabs_panel(self.f["axes-panel"], self._axes.layout(), value="axes"),
            cf.tabs_panel(self.f["legend-panel"], self._legend.layout(), value="legend"),
            cf.tabs_panel(self.f["clegend-panel"], self._clegend.layout(), value="clegend"),
            cf.tabs_panel(self.f["shape-panel"], self._shape.layout(), value="shape"),
            value="trace"
        )

    def layout(self) -> html.Div:
        return cf.div(
            self.f["div"],
            self._make_graph(),
            self._make_tabs(),
        )

    def callbacks(self) -> None:
        self._trace.callbacks()
        self._title.callbacks()
        self._axes.callbacks()
        self._legend.callbacks()
        self._clegend.callbacks()
        self._shape.callbacks()
        self._dl.callbacks()
        self._filter.callbacks()

        @callback(
            Input(self.f["rmf"], "n_clicks"),
            Output(self.fdata, "filterModel"),
            prevent_initial_call=True
        )
        def rmf(n_clicks):
            if n_clicks:
                return {}
            raise PreventUpdate

        @callback(
            Input(kid.kmviz("database"), "value"),
            Output(self.f["preset"], "data"),
            Output(self.f["preset"], "style"),
            Output(self.f["preset"], "value"),
            prevent_initial_call=True
        )
        def update_presets(db):
            if db.startswith("__kmviz_df") or self.st.mode != "db":
                raise PreventUpdate

            presets = self.st.engine.get(db).presets

            if presets.map:
                return make_select_data(list(presets.map.keys())), kshow, presets.defaults.map
            else:
                return [], khide, None

        if self.st.conf.preset == "flex":
            @callback(
                Input(self.f["preset"], "value"),
                State(kid.kmviz("database"), "value"),
                State(self._trace.f.all, "id"),
                Output(self._trace.f.all, "value"),
                State(self._title.f.all, "id"),
                Output(self._title.f.all, "value"),
                State(self._legend.f.all, "id"),
                Output(self._legend.f.all, "value"),
                prevent_initial_call=True
            )
            def update_preset_values(preset, db, ids_map, ids_title, ids_legend):
                if db and preset in self.st.engine.get(db).presets.map:
                    p = self.st.engine.get(db).presets.map[preset]
                    rmap = set_preset_values(ids_map, p.model_dump(exclude=["title", "legend"]))
                    rtitle = set_preset_values(ids_title, p.title.model_dump())
                    rlegend = set_preset_values(ids_legend, p.legend.model_dump())
                    return rmap, rtitle, rlegend
                return (no_update,) * len(ids_map), (no_update,) * len(ids_title), (no_update,) * len(ids_legend)



from kmviz.core.config import state
from kmviz.ui.components.factory import ComponentFactory as cf
from dash_extensions.enrich import html, dcc
import kmviz

class HelpLayout:
    def __init__(self, st: state, factory):
        self.st = st
        self.f = factory

    def layout(self):
        ph = []
        p_header = cf.div(self.f["plugin-header"])

        if self.st.conf.plugins:
            for name, plugin in self.st.conf.plugins.items():
                if h := plugin.help():
                    ph.append(h)
            if ph:
                p_header = cf.div(
                    self.f["plugin-header"],
                    html.H2("Plugins"),
                    html.P("The <b>kmviz</b> instance has loaded one or more plugins. You may find some help below, when provided.")
                )

        return cf.div(
            self.f["div"],
            html.H1(f"kmviz {kmviz.__version_str__}"),
            html.H2("Links"),
            html.Ul(
                [html.Li(dcc.Link("Github", href="https://github.com/tlemane/kmviz", target="_blank")),
                html.Li(dcc.Link("Documentation", href="https://tlemane.github.io/kmviz", target="_blank"))]
            ),
            *ph
        )

    def callbacks(self):
        pass



from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from dash_iconify import DashIconify
from dash_extensions.enrich import no_update, State, Input, Output, callback, clientside_callback
from dash_extensions import Keyboard
import pandas as pd
import duckdb
from dash import Patch
import dash_mantine_components as dmc

class FilterLayout:
    def __init__(self, factory, data, placeholder):
        self.f = factory
        self.data = data
        self.placeholder = placeholder

    def layout(self):
        c = [
            Keyboard(
                cf.text(
                    self.f("sql-query"),
                    placeholder=self.placeholder,
                    size="xs"),
                captureKeys=["Enter"],
                id=self.f["keyboard"],
                className="kmviz-filter-keyboard"
            ),

            dmc.Tooltip(
                label="Apply filters (pressing 'Enter' has the same effect)",
                withArrow=True,
                children=[
                    cf.action(self.f["sql-query-button"], DashIconify(icon="carbon:play-outline", width=30), variant="filled", color=km_color),
                ],
            ),
            dmc.Tooltip(
                label="Remove all filters",
                withArrow=True,
                children=[
                    cf.action(self.f["rmf"], DashIconify(icon="lucide:filter-x", width=20), variant="filled", color=km_color),
                ],
            ),
        ]

        return cf.group(
            self.f["grp-filter"],
            *c,
            gap=3
        )

    def callbacks(self):
        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return [{}, ""];
                }
                return [window.dash_clientside.no_update, window.dash_clientside.no_update];
            }
            """,
            Input(self.f["rmf"], "n_clicks"),
            Output(self.data, "filterModel"),
            Output(self.f("sql-query"), "value"),
            prevent_initial_call=True
        )

        @callback(
            Input(self.f["sql-query-button"], "n_clicks"),
            Input(self.f["keyboard"], "n_keydowns"),
            State(self.f("sql-query"), "value"),
            State(self.data, "rowData"),
            Output(self.data, "filterModel"),
            Output(self.f("sql-query"), "error"),
            prevent_initial_call=True
        )
        def df_query(n_clicks, n_keydowns, query, data):
            if n_clicks or n_keydowns:
                df = pd.DataFrame.from_dict(data)
                try:
                    if not query:
                        return {}, None

                    f = duckdb.sql(f"select ID from df WHERE {query}").df()
                    ids = list(f["ID"])
                    cond = []

                    for i in ids:
                        cond.append(
                            {"filter": i, "filterType":"text", "type":"equals"}
                        )

                    if not ids:
                        cond.append(
                            {"filter": " ", "filterType":"text", "type":"equals"}
                        )

                    p = Patch()
                    p["ID"] = { "filterType": "text", "operator": "OR", "conditions": cond}
                    return p, None
                except Exception as e:
                    return no_update, dmc.Text(str(e), color="red")
            return (no_update,) * 2


import pandas as pd
from enum import Enum
from typing import List, Union, Iterable, Dict, Any
from dataclasses import dataclass, field

from kmviz.core import KmVizError

@dataclass
class Query:
    _name: str
    _seq: str

    @property
    def name(self) -> str:
        return self._name

    @property
    def seq(self) -> str:
        return self._seq

@dataclass
class Response:
    _k: int
    _xk: float = None
    _yk: int = None
    _covxk: List[int] = None
    _covyk: List[int] = None
    _xb: float = None
    _yb: int = None
    _covxb: List[int] = None
    _covyb: List[int] = None
    _extra: Any = None

    @property
    def k(self) -> int:
        return self._k

    @property
    def xk(self) -> float:
        return self._xk

    @property
    def yk(self) -> float:
        return self._yk

    @property
    def xb(self) -> float:
        return self._xb

    @property
    def yb(self) -> float:
        return self._yb

    @property
    def covxk(self) -> List[int]:
        return self._covxk

    @property
    def covyk(self) -> List[int]:
        return self._covyk

    @property
    def covxb(self) -> List[int]:
        return self._covxb

    @property
    def covyb(self) -> List[int]:
        return self._covyb

    @property
    def extra(self) -> Any:
        return self._extra

    def has_abs(self) -> bool:
        return bool(self.yk) or bool(self.yb)


@dataclass
class QueryResponse:
    _query: Query
    _response: Dict[str, Response]
    _metadata: pd.DataFrame
    _extra: Dict[str, Any] = field(default_factory=dict)

    @property
    def query(self) -> Query:
        return self._query

    @property
    def response(self) -> Dict[str, Response]:
        return self._response

    @property
    def name(self) -> str:
        return self.query.name

    @property
    def sequence(self) -> str:
        return self.query.sequence

    @property
    def df(self) -> str:
        return self._metadata

@dataclass
class QueryResponseGeo(QueryResponse):
    _geodata: Dict[str, str] = field(default_factory=dict)


from dash_extensions.enrich import html, Input, State, Output, callback, clientside_callback, Serverside, no_update, ClientsideFunction
from dash import Patch

from kmviz.ui.utils import prevent_update_on_none, make_select_data
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, kshow, khide
from kmviz.ui.components.helpers import load_file_content
from kmviz.ui.id_factory import kid
from kmviz.core.config import state
from kmviz.core.io import KmVizIOError, parse_fastx
from kmviz.core.query import Query


import dash_mantine_components as dmc
from dash_iconify import DashIconify

class SideDBLayout:
    def __init__(self, st: state):
        self.st = st

    def layout(self) -> html.Div:
        return cf.div(
            kid.input["db-div"],
            cf.multi(
                kid.input("db"),
                data=self.st.engine.list(),
                label="Database(s)",
                clearable=True,
                withAsterisk=True,
                leftSection=DashIconify(icon="fluent:library-16-regular"),
                value=self.st.defaults.database,
            ),
            dmc.Space(h=10),
            style = khide if self.st.defaults.hide else kshow
        )

    def callbacks(self) -> None:

        @callback(
            Input(kid.input("db"), "value"),
            Output(kid.store("databases"), "data"),
            Output(kid.config["select"], "data"),
            Output(kid.config["select"], "value"),
            prevent_initial_call=True
        )
        def select_database(db_names):
            if db_names:
                return db_names, make_select_data(db_names), db_names[0]
            return [], [], None

class SideInputLayout:
    def __init__(self, st: state):
        self.st = st
        kid.register(kid.side.new("input"), "input")

    def _file_layout(self) -> html.Div:
        return cf.div(
            kid.input["file-div"],
            cf.upload(
                kid.input("file"),
                cf.button(kid.input["file-button"], "Upload FASTA/Q file", size="sm", fullWidth=True)
            ),
            style=khide
        )

    def _session_layout(self) -> html.Div:
        return cf.div(
            kid.input["session-div"],
            cf.text(kid.input("session"), required=True, label="Session ID", placeholder="kmviz-324bfabc-50e7-44ac-89d9-6460803d2f5e"),
            dmc.Space(h=2),
            cf.button(kid.input["session-button"], "Load", size="sm", fullWidth=True),
            style=khide
        )

    def _text_layout(self) -> html.Div:
        return cf.div(
            kid.input["text-div"],
            cf.area(
                kid.input("text"),
                label="Query sequence(s)",
                description="Fasta/Fastq format",
                placeholder=">Query\nACCGTAGCCTTAGAATTA",
                spellCheck=False,
                autosize=True,
                maxRows=4,
                required=True
            ),
            cf.button(kid.input["text-button"], "Load", size="xs", fullWidth=True),
            style=kshow
        )

    def layout(self) -> html.Div:
        return cf.div(
            kid.input["div"],
            cf.divider(size="sm", color="gray", label="INPUT", labelPosition="center"),
            cf.div(
                kid.input["input-div"],
                dmc.Space(h=5),
                cf.segmented(
                    kid.input["select"],
                    data=["text", "file", "session"],
                    orientation="horizontal",
                    size="sm",
                    fullWidth=True,
                    color=km_color,
                    value="text"
                ),
                dmc.Space(h=10),
                cf.div(
                    kid.input["placeholder"],
                    self._text_layout(),
                    self._file_layout(),
                    self._session_layout(),
                )
            ),
            cf.div(kid.input["error"]),
            dmc.Space(h=10)
        )

    def _text_callbacks(self) -> None:
        @callback(
            Input(kid.input["text-button"], "n_clicks"),
            State(kid.input("text"), "value"),
            Output(kid.store("sequences"), "data"),
            Output(kid.input("text"), "error"),
            Output(kid.input["input-div"], "style"),
            Output(kid.input["error"], "children"),
            prevent_initial_call=True
        )
        def load_from_text(n_clicks, content):
            if n_clicks:
                try:
                    queries = [Query(name, seq) for name, seq, _ in parse_fastx(content, self.st.limits)]
                    query = "queries" if len(queries) > 1 else "query"
                    message = dmc.Text(f"🗎 Input: {len(queries)} {query}", fw=500)
                    return Serverside(queries), no_update, khide, message
                except KmVizIOError as e:
                    return no_update, dmc.Text(str(e), fw=500, color="red"), no_update, no_update
            return no_update, no_update, no_update, no_update

    def _file_callbacks(self) -> None:
        @callback(
            Input(kid.input("file"), "filename"),
            State(kid.input("file"), "contents"),
            Output(kid.store("sequences"), "data"),
            Output(kid.input["input-div"], "style"),
            Output(kid.input["error"], "children"),
            Output(kid.input("file"), "filename"),
            prevent_initial_call=True
        )
        def load_from_file(filename, contents):
            if contents:
                content = load_file_content(filename, contents, "utf-8")
                try:
                    queries = [Query(name, seq) for name, seq, _ in parse_fastx(content, self.st.limits)]
                    query = "queries" if len(queries) > 1 else "query"
                    message = dmc.Text(f"🗎 {filename}: {len(queries)} {query}", fw=500)
                    return Serverside(queries), khide, message, None
                except KmVizIOError as e:
                    return no_update, no_update, dmc.Text(str(e), fw=500, color="red"), None
            return (no_update,) * 4

    def _session_callbacks(self) -> None:
        @callback(
            Input(kid.input["session-button"], "n_clicks"),
            State(kid.input("session"), "value"),
            Output(kid.store["results"], "data"),
            Output(kid.kmviz("database"), "data"),
            Output(kid.kmviz("database"), "value"),
            Output(kid.kmviz("query"), "data"),
            Output(kid.kmviz("query"), "value"),
            Output(kid.input("session"), "error"),
            Output(kid.store["session-id"], "data"),
            Output(kid.kmviz["side-layout"], "style"),
            Output(kid.kmviz["main-layout"], "style"),

            Output(kid.tabs["index"], "disabled"),
            Output(kid.tabs["table"], "disabled"),
            Output(kid.tabs["map"], "disabled"),
            Output(kid.tabs["plot"], "disabled"),
            Output(kid.tabs["sequence"], "disabled"),
            Output(kid.tabs["tabs"], "value"),

            prevent_initial_call=True
        )
        def load_session(n_clicks, session_id):
            if n_clicks:
                try:
                    res = self.st.get(session_id)
                    return (Serverside(res[0]), res[1], res[2], res[3], res[4], no_update, session_id, khide, {"padding-left": "10px"}) + (False,) * 5 + ("table",)
                except:
                    msg = "Session not found. Invalid session id, query still running, or results erased."
                    return (no_update,) * 5 + (msg, None) + (no_update,) * 8
            return (no_update,) * 8

    def callbacks(self) -> None:
        self._text_callbacks()
        self._file_callbacks()
        self._session_callbacks()

        clientside_callback(
            """
            function(value) {
                if (value === "text") {
                    return [{ 'display' : 'inline'}, { 'display' : 'none'}, { 'display' : 'none'}];
                } else if (value === "session") {
                    return [{ 'display' : 'none'}, { 'display' : 'inline'}, { 'display' : 'none'}];
                } else {
                    return [{ 'display' : 'none'}, { 'display' : 'none'}, { 'display' : 'inline'}];
                }
            }
            """,
            Input(kid.input["select"], "value"),
            Output(kid.input["text-div"], "style"),
            Output(kid.input["session-div"], "style"),
            Output(kid.input["file-div"], "style"),
            prevent_initial_call=True
        )



from kmviz.ui.id_factory import kid
from kmviz.core.config import state
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from dash_extensions.enrich import html, dcc
from dash_iconify import DashIconify
import dash_mantine_components as dmc
import dash_bio as dashbio
from dash.exceptions import PreventUpdate
from dash_extensions.enrich import html, Input, State, Output, callback, no_update
from kmviz.ui.utils import prevent_update_on_none, make_select_data
from kmviz.ui.layouts.figure.title import TitleLayout
from kmviz.ui.layouts.figure.axe import AxesLayout
from kmviz.ui.layouts.figure.legend import LegendLayout
from kmviz.ui.layouts.figure.shape import ShapeLayout
from kmviz.ui.layouts.figure.slider import SliderLayout
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import plotly.io as pio

import warnings
warnings.filterwarnings('ignore', module='gradpyent')

import itertools
import operator
from gradpyent import gradient

class SequenceGraphViewLayout:
    def __init__(self, st: state):
        self.st = state
        self.figure = kid.sequence.new("figure")

        self._title = TitleLayout(self.figure.new("title"), kid.sequence["figure"])
        self._axes = AxesLayout(self.figure.new("axes"), kid.sequence["figure"])
        self._legend = LegendLayout(self.figure.new("legend"), kid.sequence["figure"])
        self._shapes = ShapeLayout(self.figure.new("shape"), kid.sequence["figure"])
        self._slider = SliderLayout(self.figure.new("slider"), kid.sequence["figure"], self._axes.x.f("axis-index"))

    def _blank(self):
        fig = go.Figure(go.Scatter(x=[], y=[]))
        fig.update_layout(template="seaborn")
        fig.update_xaxes(showgrid=False, showticklabels=False, zeroline=False)
        fig.update_yaxes(showgrid=False, showticklabels=False, zeroline=False)
        return fig

    def layout(self) -> html.Div:
        return cf.div(
            kid.sequence["graph"],
            dcc.Graph(
                figure=self._blank(),
                id=kid.sequence["figure"],
                responsive=True,
                className="kmviz-dcc-sequence-graph",
            ),
            cf.tabs(
                kid.sequence["graph-tabs"],
                cf.tabs_list(
                    kid.sequence["graph-tabslist"],
                    cf.tabs_tab(kid.sequence["graph-title-tab"], "Title", "title"),
                    cf.tabs_tab(kid.sequence["graph-axes-tab"], "Axes", "axes"),
                    cf.tabs_tab(kid.sequence["graph-slider-tab"], "Slider", "slider"),
                    cf.tabs_tab(kid.sequence["graph-legend-tab"], "Legend", "legend"),
                    cf.tabs_tab(kid.sequence["graph-shape-tab"], "Shape", "shape"),
                ),
                cf.tabs_panel(kid.sequence["graph-title-panel"], self._title.layout(), "title"),
                cf.tabs_panel(kid.sequence["graph-axes-panel"], self._axes.layout(), "axes"),
                cf.tabs_panel(kid.sequence["graph-slider-panel"], self._slider.layout(), "slider"),
                cf.tabs_panel(kid.sequence["graph-legend-panel"], self._legend.layout(), "legend"),
                cf.tabs_panel(kid.sequence["graph-shape-panel"], self._shapes.layout(), "shape"),
                value="title"
            )
        )

    def callbacks(self) -> None:

        @callback(
            Input(kid.sequence["sample"], "value"),
            Input(kid.kmviz("database"), "value"),
            Input(kid.kmviz("query"), "value"),
            State(kid.store["results"], "data"),
            Output(kid.sequence["figure"], "figure"),
            Output(kid.sequence["sample"], "value"),
            prevent_initial_call=True
        )
        def update_graph_view(sample, db, query, results):
            if not db or not query or db.startswith("__kmviz_df"):
                raise PreventUpdate

            qr = results[query][db]

            if qr.df.empty:
                raise PreventUpdate

            if sample not in qr.response:
                sample = list(qr.df["ID"])[0]

            f = {"ID": {"filterType": "text", "type": "contains", "filter": sample}}

            index = list(range(len(qr.query.seq)))

            m = []
            cols = ["base"]

            if qr.response[sample].k:
                m = [None] * (qr.response[sample].k - 1)
                cols = ["kmer", "base"]

            if qr.response[sample].has_abs():
                if qr.response[sample].covyk is None:
                    raise PreventUpdate
                df = pd.DataFrame({"Sequence": index, "kmer": qr.response[sample].covyk + m, "base": qr.response[sample].covyb})
                fig = px.line(df, x="Sequence", y=cols, line_shape="linear", markers=True)
            else:
                if qr.response[sample].covxk is None:
                    raise PreventUpdate
                df = pd.DataFrame({"Sequence": index, "kmer": qr.response[sample].covxk + m, "base": qr.response[sample].covxb})
                fig = px.line(df, x="Sequence", y=cols, line_shape="linear", markers=True)

            fig.update_layout(
                xaxis = dict(
                    tickmode = "array",
                    tickvals = index,
                    ticktext = [x for x in qr.query.seq]
                )
            )

            fig.update_layout(
                xaxis = dict(rangeslider=dict(visible=True, range=[0, len(qr.query.seq)]),
                             range=[0, 50],
                             type="category")
            )

            fig.update_layout(legend_title_text="")
            fig.update_xaxes(title_text="Sequence")
            fig.update_yaxes(rangemode="tozero")

            return fig, sample

        self._title.callbacks()
        self._axes.callbacks()
        self._legend.callbacks()
        self._shapes.callbacks()
        self._slider.callbacks()

class SequenceTextViewLayout:
    def __init__(self, st: state):
        self.st = st

    def _make_coverage(self, cov: list, abs, start, end):
        grad = gradient.Gradient(gradient_start=start, gradient_end=end)

        ranges = []
        series = []

        coverage = []
        legend = []

        for r in itertools.groupby(enumerate(cov), key=operator.itemgetter(1)):
            series.append(r[0])
            L = list(r[1])
            ranges.append((L[0][0], L[-1][0] + 1))

        colors = grad.get_gradient_series(series, fmt="html")

        for i, c in enumerate(colors):
            coverage.append({
                "bgcolor": c,
                "start": ranges[i][0],
                "end": ranges[i][1],
                "tooltip": round(series[i], 2),
                "onclick": round(series[i], 2),
            })

        if all(isinstance(a, int) for a in series):
            sorted_series = sorted(zip(series, colors))

            s = set()
            for value, color in sorted_series:
                if value not in s:
                    legend.append({
                        "color": color,
                        "name": value
                    })
                    s.add(value)

        else:
            legend.append({
                "color": colors[min((val, idx) for (idx, val) in enumerate(series))[1]],
                "name": round(min(series), 2)
            })
            legend.append({
                "color": colors[max((val, idx) for (idx, val) in enumerate(series))[1]],
                "name": round(max(series), 2)
            })

        return coverage, legend

    def layout(self) -> html.Div:
        return cf.div(
            kid.sequence["text"],
            dmc.Space(h=10),
            cf.group(
                kid.sequence["text-grp"],
                cf.segmented(
                    kid.sequence["text-mode"],
                    data=["base", "kmer"],
                    value="base",
                    color=km_color,
                    size="xs",
                    className = "kmviz-figure-segmented"
                ),
                cf.color(
                    kid.sequence["start"],
                    label="Min color",
                    size="xs",
                    leftSection=DashIconify(icon="cil:paint"),
                    withPreview=False,
                    value="#E0DFDF",
                    noTranspa=True,
                ),
                cf.color(
                    kid.sequence["end"],
                    label="Max color",
                    size="xs",
                    leftSection=DashIconify(icon="cil:paint"),
                    withPreview=False,
                    value=km_color,
                    noTranspa=True,
                )
            ),

            cf.div(
                kid.sequence["view-div"],
                dashbio.SequenceViewer(
                    id=kid.sequence["view"],
                    wrapAminoAcids=True,
                    toolbar=False,
                    search=False,
                    badge=False,
                    charsPerLine=130,
                    sequenceMaxHeight="55vh"
                ),
                style = { "margin-left": "auto", "margin-right": "auto", "position": "relative"}
            )
        )

    def callbacks(self) -> None:

        @callback(
            Input(kid.sequence["sample"], "value"),
            Input(kid.kmviz("database"), "value"),
            Input(kid.kmviz("query"), "value"),
            Input(kid.sequence["text-mode"], "value"),
            Input(kid.sequence["start"], "value"),
            Input(kid.sequence["end"], "value"),
            State(kid.store["results"], "data"),
            Output(kid.sequence["view"], "sequence"),
            Output(kid.sequence["view"], "coverage"),
            Output(kid.sequence["view"], "legend"),
            Output(kid.sequence["sample"], "value"),
            prevent_initial_call=True
        )
        def update_text_view(sample, db, query, mode, start, end, results):
            if not db or not query or db.startswith("__kmviz_df"):
                raise PreventUpdate
            prevent_update_on_none(sample, db, query)

            qr = results[query][db]

            if qr.df.empty:
                raise PreventUpdate

            if sample not in qr.response:
                sample = list(qr.df["ID"])[0]

            if mode == "kmer" and not qr.response[sample].k:
                mode = "base"

            if qr.response[sample].has_abs():
                if qr.response[sample].covyk is None:
                    raise PreventUpdate
                cov, leg = self._make_coverage(qr.response[sample].covyb if mode == "base" else qr.response[sample].covyk, True, start, end)
            else:
                if qr.response[sample].covxk is None:
                    raise PreventUpdate
                cov, leg = self._make_coverage(qr.response[sample].covxb if mode == "base" else qr.response[sample].covxk, True, start, end)

            return qr.query.seq, cov, leg, sample

class SequenceLayout:
    def __init__(self, st: state):
        self.st = st
        self._text = SequenceTextViewLayout(self.st)
        self._graph = SequenceGraphViewLayout(self.st)

    def layout(self):
        return cf.tabs(
                kid.sequence["tabs"],
                cf.tabs_list(
                    kid.sequence["tabslist"],
                    cf.tabs_tab(kid.sequence["graph-view"], "Graph view", value="graph", leftSection=DashIconify(icon="mdi:graph-line")),
                    cf.tabs_tab(kid.sequence["text-view"], "Text view", value="text", leftSection=DashIconify(icon="mdi:dna")),
                    dmc.Space(w=10),
                    cf.group(
                        kid.sequence["tab-head-grp"],
                        cf.select(
                            kid.sequence["sample"],
                            [],
                            size="xs",
                            leftSection=DashIconify(icon="mingcute:document-fill"),
                            className="kmviz-sample-select"
                        ),
                        cf.action(kid.sequence["download-button"], DashIconify(icon="bi:filetype-json", width=20), variant="filled", color=km_color),
                        dcc.Download(id=kid.sequence["download"])
                    )
                ),
                cf.tabs_panel(kid.sequence["graph-view-panel"], self._graph.layout(), value="graph"),
                cf.tabs_panel(kid.sequence["text-view-panel"], self._text.layout(), value="text"),
                value = "graph"
        )

    def callbacks(self) -> None:
        self._text.callbacks()
        self._graph.callbacks()

        @callback(
            Input(kid.kmviz("database"), "value"),
            Input(kid.kmviz("query"), "value"),
            State(kid.store["results"], "data"),
            Output(kid.sequence["sample"], "data"),
            Output(kid.sequence["sample"], "value"),
            prevent_initial_call=True
        )
        def update_sequence_data(db, query, results):
            if not db or not query or db.startswith("__kmviz_df"):
                prevent_update_on_none(None)

            qr = results[query][db]
            ids = list(qr.df["ID"])
            if not ids:
                return [], None
            return make_select_data(ids), ids[0]

        @callback(
            Input(kid.sequence["download-button"], "n_clicks"),
            State(kid.kmviz("database"), "value"),
            State(kid.kmviz("query"), "value"),
            State(kid.store["results"], "data"),
            Output(kid.sequence["download"], "data"),
            prevent_initial_call=True
        )
        def download_json(n_clicks, db, query, results):
            if n_clicks:
                if not db or not query or db.startswith("__kmviz_df"):
                    raise PreventUpdate

                qr = results[query][db]

                res={query: {}}

                for sample, response in qr.response.items():
                    if response.has_abs():
                        res[query][sample] = {"ykmer": response.covyk, "ybase": response.covyb}
                    else:
                        res[query][sample] = {"xkmer": response.covxk, "xbase": response.covxb}

                return dict(content=str(res), filename=f"{db}-{query}.json")
            return no_update





from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from kmviz.ui.utils import icons
from kmviz.ui.id_factory import IDFactory, kid
from dash_extensions.enrich import Input, Output, no_update, callback, ctx
from kmviz.ui.layouts.figure import patch_figure, patch_figure_all
import dash_mantine_components as dmc

class LegendLayout:
    def __init__(self, factory: IDFactory, figure_id):
        self.f = factory
        self.fid = figure_id

    def _title_panel(self):
        return cf.tabs_panel(
            self.f["title-panel"],
            [
                cf.group(
                    self.f["title-grp-1"],
                    cf.text(self.f("legend_title_text"), label="Text", placeholder="<b>Title</b>", size="xs"),
                    cf.number(self.f("legend_title_font_size"), label="Font size", leftSection=icons("font_size"), min=0, max=100, value=None, step=1, placeholder="14", size="xs"),
                    cf.font(self.f("legend_title_font_family"), label="Font", size="xs"),
                    cf.select(self.f("legend_title_side"), data=["top", "left", "top left", "top center", "top right"], label="Position", value="top", size="xs", leftSection=icons("pos"), className="kmviz-figure-select"),
                    cf.color(self.f("legend_title_font_color"), label="Font color", size="xs", value="#000000"),
                )
            ],
            value="title"
        )

    def _position_panel(self):
        orient = [
            {"label": "vertical", "value": "v"},
            {"label": "horizontal", "value": "h"},
        ]
        return cf.tabs_panel(
            self.f["position-panel"],
            [
                cf.group(
                    self.f["position-grp-1"],
                    cf.select(self.f("legend_xanchor"), label="Anchor X", data=["auto", "left", "center", "right"], value="auto", leftSection=icons("single"), size="xs"),
                    cf.select(self.f("legend_xref"), label="Reference X", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                    cf.number(self.f("legend_x"), label="x", min=0.0, max=1.0, step=0.01, value=0.9, startValue=0.9, placeholder="0.5", decimalScale=2, leftSection=icons("floating"), size="xs"),
                    cf.select(self.f("legend_yanchor"), label="Anchor Y", data=["auto", "top", "middle", "bottom"], value="auto", leftSection=icons("single"), size="xs"),
                    cf.select(self.f("legend_yref"), label="Reference Y", data=["container", "paper"], value="container", leftSection=icons("single"), size="xs"),
                    cf.number(self.f("legend_y"), label="y", min=0.0, max=1.0, step=0.01, value=None, startValue=0.9, placeholder="0.95", decimalScale=2, leftSection=icons("floating"), size="xs"),
                    cf.number(self.f("legend_indentation"), label="Identation", min=-15, max=100, step=1, value=0, leftSection=icons("indent"), size="xs"),
                    cf.segmented(self.f("legend_orientation"), data=orient, value="v", size="xs", className="kmviz-figure-segmented"),
                )
            ],
            value="position"
        )

    def _box_panel(self):
        return cf.tabs_panel(
            self.f["box-panel"],
            [
                cf.group(
                    self.f["box-grp-1"],
                    cf.number(self.f("legend_borderwidth"), label="Width", min=0, max=100, step=1, value=0, leftSection=icons("width", rotate=1), size="xs"),
                    cf.color(self.f("legend_bordercolor"), label="Border color", size="xs"),
                    cf.color(self.f("legend_bgcolor"), label="Bg color", size="xs"),
                )
            ],
            value="box"
        )

    def _entry_panel(self):
        return cf.tabs_panel(
            self.f["entry-panel"],
            [
                cf.group(
                    self.f["entry-grp-1"],
                    cf.font(self.f("legend_font_family"), label="Font", size="xs", className="kmviz-figure-select"),
                    cf.number(self.f("legend_font_size"), label="Font size", leftSection=icons("font_size"), min=0, max=100, value=None, step=1, placeholder="10", size="xs"),
                    cf.color(self.f("legend_font_color"), label="Font color", size="xs", value="#000000")
                )
            ],
            value="entry"
        )

    def layout(self):
        return cf.tabs(
            self.f["legend"],
            cf.tabs_list(
                self.f["legend-tabslist"],
                cf.tabs_tab(self.f["title-tab"], "Title", value="title"),
                cf.tabs_tab(self.f["position-tab"], "Position", value="position"),
                cf.tabs_tab(self.f["entry-tab"], "Entry", value="entry"),
                cf.tabs_tab(self.f["box-tab"], "Box", value="box"),
                dmc.Space(w=10),
                cf.switch(self.f("showlegend"), checked=True, size="lg", offLabel="Show", onLabel="Hide", style = {"margin-top": "2px"}, color=km_color),
            ),
            self._title_panel(),
            self._position_panel(),
            self._entry_panel(),
            self._box_panel(),
            value=None
        )

    def callbacks(self):

        @callback(
            Output(self.fid, "figure", allow_duplicate=True),
            inputs=dict(inputs=(
                Input(self.fid, "figure"),
                Input(kid.kmviz("auto"), "checked"),
                Input(self.f("showlegend"), "checked"),
                Input(self.f("legend_title_text"), "value"),
                Input(self.f("legend_title_font_size"), "value"),
                Input(self.f("legend_title_font_family"), "value"),
                Input(self.f("legend_title_font_color"), "value"),
                Input(self.f("legend_title_side"), "value"),
                Input(self.f("legend_font_size"), "value"),
                Input(self.f("legend_font_family"), "value"),
                Input(self.f("legend_font_color"), "value"),
                Input(self.f("legend_borderwidth"), "value"),
                Input(self.f("legend_bordercolor"), "value"),
                Input(self.f("legend_bgcolor"), "value"),
                Input(self.f("legend_indentation"), "value"),
                Input(self.f("legend_orientation"), "value"),
                Input(self.f("legend_xanchor"), "value"),
                Input(self.f("legend_xref"), "value"),
                Input(self.f("legend_x"), "value"),
                Input(self.f("legend_yanchor"), "value"),
                Input(self.f("legend_yref"), "value"),
                Input(self.f("legend_y"), "value"),
            )),
            prevent_initial_call=True,
            prevent_initial_callbacks= True
        )
        def update_legend(inputs):
            if ctx.triggered_id == self.fid:
                if not inputs[1]:
                    return no_update
                return patch_figure_all(ctx, ["layout"], set([self.fid, kid.kmviz("auto", True)]))
            return patch_figure(ctx)


from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.utils import icons
from dash_extensions.enrich import Input, State, Output, callback, ctx
from kmviz.ui.layouts.figure import patch_figure

class SliderLayout:
    def __init__(self, factory, figure_id, axis):
        self.f = factory
        self.fid = figure_id
        self.axis = axis

    def layout(self):
        return cf.group(
            self.f["slider-grp-1"],
            cf.switch(self.f("rangeslider_visible"), onLabel="ON", offLabel="OFF", checked=False, size="lg", className="kmviz-figure-switch"),
            cf.number(self.f("rangeslider_borderwidth"), min=0, value=1, max=100, step=1, leftSection=icons("cline"), label="Width", size="xs"),
            cf.number(self.f("rangeslider_thickness"), min=0.0, value=0.1, max=1, step=0.01, decimalScale=2, leftSection=icons("floating"), label="Thickness", size="xs"),
            cf.segmented(self.f("rangeslider_yaxis_rangemode"), ["auto", "fixed", "match"], value="auto", className="kmviz-figure-segmented", size="xs"),
            cf.color(self.f("rangeslider_bordercolor"), size="xs", label="Border color", value=None),
            cf.color(self.f("rangeslider_bgcolor"), size="xs", label="Bg color", value=None),
        )

    def callbacks(self):
        @callback(
            Output(self.fid, "figure"),
            inputs=dict(inputs=(
                State(self.axis, "value"),
                Input(self.f("rangeslider_visible"), "checked"),
                Input(self.f("rangeslider_borderwidth"), "value"),
                Input(self.f("rangeslider_thickness"), "value"),
                Input(self.f("rangeslider_bordercolor"), "value"),
                Input(self.f("rangeslider_bgcolor"), "value"),
                Input(self.f("rangeslider_yaxis_rangemode"), "value"),
            )),
            prevent_initial_call=True,
        )
        def update_rs(inputs):
            n = "xaxis" if inputs[0] == 0 else f"xaxis{inputs[0]+1}"
            return patch_figure(ctx, ["layout", n])



from .notifier import Notifier, NullNotifier
from kmviz.core.notifier.sendgrid import SendGridNotifier
from kmviz.core.notifier.smtp import SMTPNotifier

NOTIFIERS = {
    "sendgrid": SendGridNotifier,
    "smtp": SMTPNotifier
}




from abc import ABC, abstractmethod
from kmviz.core.query import Query
from kmviz.core.log import kmv_info, kmv_warn
from kmviz.core.provider.options import TextOption

class Notifier(ABC):
    def __init__(self, success: str, failure: str, subject: str, subject_failure: str, custom: dict = {}):
        self._success = success
        self._failure = failure
        self._subject = subject
        self._subject_failure = subject_failure
        self._custom = custom

    @abstractmethod
    def send_success(self, session: str, options: dict, results: dict = {}):
        """
        Send success notification
        """

    @abstractmethod
    def send_failure(self, session: str, options: dict, reason: str = None):
        """
        Send failure notification
        """

    def check(self, options: dict):
        pass

    def format_success(self, session: str, **kwargs):
        return self._success.format(SESSION=session, **kwargs, **self._custom)

    def format_failure(self, session: str, reason: str, **kwargs):
        return self._failure.format(SESSION=session, REASON=reason, **kwargs, **self._custom)

    def format_subject(self, session: str, **kwargs):
        return self._subject.format(SESSION=session, **kwargs, **self._custom)

    def format_subject_failure(self, session: str, reason: str, **kwargs):
        return self._subject_failure.format(SESSION=session, REASON=reason, **kwargs, **self._custom)

    def options(self):
        return {}

class NullNotifier(Notifier):
    def __init__(self):
        super().__init__("", "", "", "")

    def send_success(self, query: Query, session: str, options: dict):
        kmv_info(f"[notif-success] {session}")

    def send_failure(self, query: Query, session: str, options: dict, reason: str = None):
        kmv_warn(f"[notif-failure] {session} (reason: {str(reason)})")


import re

def is_email(s: str):
    return re.match(r'^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$', s)



from pydantic import BaseModel
from typing import Literal, Optional, Annotated, Union, List, Dict, Any
from annotated_types import Ge, Le
from dash_extensions.enrich import no_update

class title_presets(BaseModel):
    title_text: Optional[str]=None
    title_font_size: Optional[int]=None
    title_font_familiy: Optional[Literal["Arial", "Balto", "Courier New", "Droid Sans", "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman"]]=None
    title_font_color: Optional[str]=None

    title_xanchor: Optional[Literal["auto", "left", "center", "right"]]=None
    title_xref: Optional[Literal["container", "paper"]]=None
    title_x: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None

    title_yanchor: Optional[Literal["auto", "top", "middle", "bottom"]]=None
    title_yref: Optional[Literal["container", "paper"]]=None
    title_y: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None

class legend_presets(BaseModel):
    legend_title_text: Optional[str]=None
    legend_title_font_size: Optional[int]=None
    legend_title_font_familiy: Optional[Literal["Arial", "Balto", "Courier New", "Droid Sans", "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman"]]=None
    legend_title_font_color: Optional[str]=None

    legend_xanchor: Optional[Literal["auto", "left", "center", "right"]]=None
    legend_xref: Optional[Literal["container", "paper"]]=None
    legend_x: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None
    legend_yanchor: Optional[Literal["auto", "top", "middle", "bottom"]]=None
    legend_yref: Optional[Literal["container", "paper"]]=None
    legend_y: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None

    legend_indentation: Annotated[Optional[int], Ge(-15), Le(100)]=None
    legend_orientation: Optional[Literal["v", "h"]]=None

    legend_font_size: Optional[int]=None
    legend_font_familiy: Optional[Literal["Arial", "Balto", "Courier New", "Droid Sans", "Droid Serif", "Droid Sans Mono", "Gravitas One", "Old Standard TT", "Open Sans", "Overpass", "PT Sans Narrow", "Raleway", "Times New Roman"]]=None
    legend_font_color: Optional[str]=None

    legend_borderwidth: Annotated[Optional[int], Ge(0), Le(100)]=None
    legend_bordercolor: Optional[str]=None
    legend_bgcolor: Optional[str]=None

class map_presets(BaseModel):
    color: Optional[str]=None
    size: Optional[str]=None
    text: Optional[str]=None
    symbol: Optional[str]=None
    animation_frame: Optional[str]=None
    animation_group: Optional[str]=None
    template: Optional[str]=None
    projection: Optional[str]=None
    color_seq_continuous_scale: Optional[str]=None
    color_div_continuous_scale: Optional[str]=None
    color_cyc_continuous_scale: Optional[str]=None
    color_continuous_midpoint: Optional[Union[int, float]]=None
    opacity: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None
    size_max: Annotated[Optional[int], Ge(0), Le(50)]=None
    size_min: Annotated[Optional[int], Ge(0), Le(50)]=None
    color_discrete_map: Optional[Dict[Any, str]]=None
    symbol_map: Optional[Dict[Any, str]]=None
    color_discrete_sequence: Optional[List[str]]=None
    symbol_sequence: Optional[List[str]]=None

    title: Optional[title_presets]=title_presets()
    legend: Optional[legend_presets]=legend_presets()

class plot_presets(BaseModel):
    type: str

    X: Optional[List[str]]=None
    Y: Optional[List[str]]=None
    Z: Optional[List[str]]=None
    color: Optional[str]=None
    size: Optional[str]=None
    text: Optional[str]=None
    symbol: Optional[str]=None
    pattern_shape: Optional[str]=None
    base: Optional[str]=None
    line_dash: Optional[str]=None
    line_group: Optional[str]=None
    dimensions: Optional[List[str]]=None
    values: Optional[str]=None
    names: Optional[str]=None

    animation_frame: Optional[str]=None
    animation_group: Optional[str]=None
    template: Optional[str]=None

    color_seq_continuous_scale: Optional[str]=None
    color_div_continuous_scale: Optional[str]=None
    color_cyc_continuous_scale: Optional[str]=None
    color_continuous_midpoint: Optional[Union[int, float]]=None
    opacity: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None
    size_max: Annotated[Optional[int], Ge(0), Le(50)]=None
    size_min: Annotated[Optional[int], Ge(0), Le(50)]=None

    color_discrete_map: Optional[Dict[Any, str]]=None
    symbol_map: Optional[Dict[Any, str]]=None
    color_discrete_sequence: Optional[List[str]]=None
    symbol_sequence: Optional[List[str]]=None

    pattern_shape_map: Optional[Dict[Any, str]]=None
    line_dash_map: Optional[Dict[Any, str]]=None
    pattern_shape_sequence: Optional[List[str]]=None
    line_dash_sequence: Optional[List[str]]=None

    trendline: Optional[Literal["ols", "lowess", "rolling", "expanding", "ewm"]]=None
    trendline_scope: Optional[Literal["trace", "overall"]]=None
    trendline_options: Optional[Dict[Any, str]]=None

    marginal_x: Optional[Literal["rug", "box", "violin", "histogram"]]=None
    marginal_y: Optional[Literal["rug", "box", "violin", "histogram"]]=None

    histfunc: Optional[Literal["count", "sum", "avg", "min", "max"]]=None
    histnorm: Optional[Literal["rug", "box", "violin", "histogram"]]=None
    nbins: Annotated[Optional[int], Ge(1), Le(1000)]=None

    facet_row: Optional[str]=None
    facet_col: Optional[str]=None
    facet_col_spacing: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None
    facet_row_spacing: Annotated[Optional[float], Ge(0.0), Le(1.0)]=None
    facet_col_wrap: Annotated[Optional[int], Ge(0), Le(10)]=None

    title: Optional[title_presets]=title_presets()
    legend: Optional[legend_presets]=legend_presets()

class cpresets_defaults(BaseModel):
    map: Optional[str]=None
    plot: Optional[str]=None

class cpresets(BaseModel):
    priority: Optional[bool]=None
    map: Optional[Dict[str, map_presets]]=None
    plot: Optional[Dict[str, plot_presets]]=None
    defaults: Optional[cpresets_defaults]=cpresets_defaults()

def set_preset_values(ids: list, preset: dict) -> list:
    r = [no_update for _ in range(len(ids))]
    for index, id in enumerate(ids):
        if id["index"] in ("xselect", "yselect", "zselect"):
            id["index"] = id["index"][0].upper()
        if id["index"] in preset:
            r[index] = preset[id["index"]]
    return r




from kmviz.core.config import state
import kmviz.core.config as kconf
from dash_extensions.enrich import html, callback, Serverside, Input, Output, State, no_update
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color, khide, kshow
from kmviz.ui.id_factory import kid
from kmviz.ui.utils import make_select_data
from kmviz.core.log import kmv_info, kmv_warn, kmv_error
from dash_iconify import DashIconify
import dash_mantine_components as dmc
from kmviz.core import KmVizQueryError, KmVizQueryErrorNotif
import uuid

class SubmitLayout:
    def __init__(self, st: state):
        self.st = st
        self._def_tab = "table"


    def _make_notif_msg(self, current: int, nb_queries: int) -> str:
        return f"[{current}/{nb_queries}]"

    def _update_submit_notification(self, title, message):
        return dmc.Alert(
            f"Processing... {message} {self.st.ui.ui_notif_msg}",
            title=title,
            icon=DashIconify(icon="eos-icons:hourglass", width=30),
            color=km_color,
            withCloseButton=True,
            className = "kmviz-alert",
            variant="outline"
        )

    def _update_ok_submit_notification(self, title, message):
        return dmc.Alert(
            "",
            title=title if not message else f"{title} {message}",
            icon=DashIconify(icon="ep:success-filled", width=30),
            color=km_color,
            withCloseButton=True,
            className = "kmviz-alert",
            variant="outline"
        )

    def _make_error_submit_notification(self, title, msg = None):
        message = "An error occurred while processing your request(s)"

        if msg:
            message = msg

        return dmc.Alert(
            message,
            title=title,
            color="red",
            icon=DashIconify(icon="ic:twotone-error", width=30),
            withCloseButton=True,
            className = "kmviz-alert",
            variant="outline"
        )

    def _on_error(self, uuid_str, msg=None):
        return (no_update,) * 6 + (self._make_error_submit_notification(uuid_str, msg), False) + (no_update,) * 2 + (True,) * 5 + (self._def_tab, kshow, [])

    def _submit_callback(self):

        @callback(
            Input(kid.store["state"], "data"),
            State(kid.submit["button"], "n_clicks"),
            Output(kid.submit["button"], "disabled"),
            prevent_initial_call=True
        )
        def is_ready(state, n_clicks):
            #if n_clicks and n_clicks > 0:
            #    return True
            if all(state.values()):
                return False
            return True

        @callback(
            Output(kid.store["results"], "data"),
            Output(kid.kmviz("database"), "data"),
            Output(kid.kmviz("database"), "value"),
            Output(kid.kmviz("query"), "data"),
            Output(kid.kmviz("query"), "value"),
            Output(kid.store["session-id"], "data"),
            Output(kid.submit["notification"], "children"),
            Output(kid.submit["button"], "disabled"),
            Output(kid.kmviz["side-layout"], "style"),
            Output(kid.kmviz["main-layout"], "style"),

            Output(kid.tabs["index"], "disabled"),
            Output(kid.tabs["table"], "disabled"),
            Output(kid.tabs["map"], "disabled"),
            Output(kid.tabs["plot"], "disabled"),
            Output(kid.tabs["sequence"], "disabled"),
            Output(kid.tabs["tabs"], "value"),

            Output(kid.input["input-div"], "style"),
            Output(kid.input["error"], "children"),

            inputs=[
                Input(kid.submit["button"], "n_clicks"),
                State(kid.store("sequences"), "data"),
                State(kid.store("options"), "data"),
                State(kid.input("db"), "value")
            ],
            background=True,
            running=[
                (Output(kid.submit["button"], "disabled"), False, True),
                (Output(kid.submit["button"], "loading"), True, False)
            ],
            progress=[
                Output(kid.submit["notification"], "children"),
            ],
            prevent_initial_call=True
        )
        def submit(set_progress, n_clicks, sequences, options, actives):
            uuid_str = f"kmviz-{str(uuid.uuid4())}"

            nb_queries = len(sequences)

            kmv_info(f"⌛ {uuid_str}")

            results = {}

            set_progress((self._update_submit_notification(uuid_str, self._make_notif_msg(0, nb_queries))))

            try:
                kconf.st.notif.check(options["notif"])
                for i, query in enumerate(sequences):
                    result = kconf.st.engine.query(query, actives, options, uuid_str)
                    keys = list(result.keys())
                    for key in keys:
                        if isinstance(result[key], str):
                            raise KmVizQueryError(result[key])
                    results[query.name] = result
                    set_progress((self._update_submit_notification(uuid_str, self._make_notif_msg(i+1, nb_queries))))
            except KmVizQueryError as e:
                kmv_warn(f"⚠️  {uuid_str} -> {str(e)}")
                return self._on_error(uuid_str, str(e))
            except KmVizQueryErrorNotif as e:
                kmv_warn(f"⚠️  {uuid_str} -> {str(e)}")
                kconf.st.notif.send_failure(uuid_str, options["notif"], str(e))
                return self._on_error(uuid_str, str(e))
            except Exception as e:
                kmv_warn(f"⚠️  {uuid_str} -> {str(e)}")
                return self._on_error(uuid_str)

            def_query = sequences[0].name
            def_db = actives[0]

            data_db = make_select_data(actives)
            data_query = [query.name for query in sequences]

            store = (results, data_db, def_db, data_query, def_query)

            kconf.st.put(uuid_str, store)

            kmv_info(f"✅ {uuid_str}")

            notif = self._update_ok_submit_notification(f"{uuid_str}", None),

            kconf.st.notif.send_success(uuid_str, options["notif"], results)

            return (Serverside(store[0]), *store[1:], uuid_str, notif, True) + (khide, {"padding-left": "10px"}) + (False,) * 5 + (self._def_tab, khide, no_update)

    def layout(self) -> html.Div:
        return cf.div(
            kid.submit["div"],
            dmc.Space(h=10),
            dmc.Center([
                cf.group(
                    kid.submit["grp-button"],
                    cf.button(
                        kid.submit["button"],
                        "Submit",
                        leftSection=DashIconify(icon="mdi:dna", width=20),
                        disabled=True,
                        loaderProps={"type": "dots"}
                    ),
                    html.A(
                        cf.button(
                            kid.submit["reset"],
                            "Reset",
                            leftSection=DashIconify(icon="system-uicons:refresh", width=20),
                        ),
                        href=self.st.instance_plugin[0]
                    )
                )
            ]),
        )

    def callbacks(self) -> None:
        self._submit_callback()


from kmviz.core.config import state
from dash_extensions.enrich import html, Input, State, callback, Output, no_update, clientside_callback, dcc
from kmviz.ui.components.factory import ComponentFactory as cf
from kmviz.ui.components.factory import km_color
from kmviz.ui.id_factory import kid
import dash_mantine_components as dmc
from dash_iconify import DashIconify
from kmviz.ui.utils import prevent_update_on_none
from dash.exceptions import PreventUpdate
from dash import Patch
import duckdb
import pandas as pd
from pandas.api.types import is_numeric_dtype, is_datetime64_dtype
from kmviz.ui.layouts.filter import FilterLayout
import uuid

class TableLayout:
    def __init__(self, st: state):
        self.st = st
        self._filter = FilterLayout(kid.table, kid.table("grid"), st.ui.table_filter_message)

    def layout(self) -> html.Div:
        return cf.div(
            kid.table["div"],
            dmc.Space(h=5),
            self._filter.layout(),
            dmc.Space(h=5),
            dcc.Loading(id=kid.table("loading"), type="default", delay_show=250, children=[
                cf.ag_grid(kid.table("grid"), {}, {}, {}, style = {"height": "80vh"}),
            ]),
            dmc.Space(h=5),
            cf.group(
                kid.table["grp-button"],
                cf.button(
                    kid.table["export-button"],
                    "Export",
                    disabled=True,
                    leftSection=DashIconify(icon="ph:export", width=20)
                ),
                cf.button(
                    kid.table["rmf-button"],
                    "Remove filters",
                    disabled=True,
                    leftSection=DashIconify(icon="ph:trash", width=20)
                ),
                cf.button(
                    kid.table["filter-button"],
                    "Filter NaN",
                    disabled=True,
                    leftSection=DashIconify(icon="mingcute:na-fill", width=20)
                )
            ),
        )

    def callbacks(self) -> None:
        clientside_callback(
            """
            function(database, query) {
                if (!database || !query) {
                    return [window.dash_clientside.no_update, window.dash_clientside.no_update, window.dash_clientside.no_update];
                }
                return [false, false, false];
            }
            """,
            Input(kid.kmviz("query"), "value"),
            State(kid.kmviz("database"), "value"),
            Output(kid.table["filter-button"], "disabled"),
            Output(kid.table["export-button"], "disabled"),
            Output(kid.table["rmf-button"], "disabled"),
            prevent_initial_call=True
        )

        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return {};
                }
                return window.dash_clientside.no_update;
            }
            """,
            Input(kid.table["rmf-button"], "n_clicks"),
            Output(kid.table("grid"), "filterModel"),
            prevent_initial_call=True
        )

        clientside_callback(
            """
            function(n_clicks) {
                if (n_clicks) {
                    return true;
                }
                return window.dash_clientside.no_update;
            }
            """,
            Input(kid.table["export-button"], "n_clicks"),
            Output(kid.table("grid"), "exportDataAsCsv"),
            prevent_initial_call=True
        )

        @callback(
            Input(kid.table["filter-button"], "n_clicks"),
            State(kid.table("grid"), "columnDefs"),
            Output(kid.table("grid"), "filterModel", allow_duplicate=True),
            prevent_initial_call=True
        )
        def filter_nan(n_clicks, fields):
            if n_clicks:
                cols = [x["field"] for x in fields]
                p = Patch()
                for c in cols[1:]:
                    p[c] = {'filterType': 'text', 'type': 'notBlank', 'filter': ''}
                return p

        @callback(
            Input(kid.kmviz("query"), "value"),
            Input(kid.pom["store"], "data"),
            State(kid.kmviz("database"), "value"),
            State(kid.store["results"], "data"),
            Output(kid.table("grid"), "rowData"),
            Output(kid.table("grid"), "columnDefs"),
            prevent_initial_call=True
        )
        def update_grid(query, plot_only, db, results):

            def colf(data):
                if is_numeric_dtype(data):
                    return "agNumberColumnFilter"
                if is_datetime64_dtype(data):
                    return "agDateColumnFilter"
                return "agTextColumnFilter"


            if "df" not in plot_only:
                df = results[query][db].df
            else:
                df = plot_only["df"]

            fields = [
                {
                    "field": x,
                    "headerName": x,
                    "filterParams": {"maxNumConditions": 10000},
                    "suppressMenu": True,
                    "filter": colf(df[x]),
                    "cellRenderer": self.st.ui.crs[x][0] if x in self.st.ui.crs else "",
                    "cellRendererParams": self.st.ui.crs[x][1] if x in self.st.ui.crs else ""
                } for x in list(df)
            ]

            return df.to_dict("records"), fields

        self._filter.callbacks()


from dash_extensions.enrich import no_update
from dash import Patch
from kmviz.ui.utils import patch_value, patch_value_id
from kmviz.ui.components.factory import dmc_new

def color_to_string(color):
    return color

def patch_figure(context, paths = ["layout"]):
    args = context.args_grouping
    trigger = context.triggered_id

    for e in args["inputs"]:
        if e["id"] == trigger:
            if isinstance(e["value"], dict):
                e["value"] = color_to_string(e["value"])
            return patch_value_id(Patch(paths), trigger, e["value"])

    return no_update

def patch_figure_all(context, paths = ["layout"], skip = set()):
    args = context.args_grouping
    p = Patch(paths)
    for e in args["inputs"]:
        if e["str_id"] not in skip:
            if (isinstance(e["value"], list) and not e["value"]) or e["value"] is None:
                continue
            if isinstance(e["value"], dict):
                e["value"] = color_to_string(e["value"])
            patch_value_id(p, e["id"], e["value"])
    return p

def apply_presets(presets: dict, params: dict, priority: bool=True):
    if not presets:
        return params

    for k in list(presets.keys()):
        if presets[k] is None:
            del presets[k]

    if priority:
        return {**params, **presets}
    else:
        for k, v in params.items():
            if k in presets and v is not None:
                presets[k] = v
        return {**params, **presets}

def apply_legend_and_title_presets(fig, presets):
    title = {k: v for k, v in presets.title.model_dump().items() if v is not None}
    legend = {k: v for k, v in presets.legend.model_dump().items() if v is not None}

    if len(title):
        fig.update_layout(**title)

    if len(legend):
        fig.update_layout(**legend)




from kmviz.core.provider import Provider
from kmviz.core.metadata.db import MetaDB
from kmviz.core.notifier import Notifier
from kmviz.core import KmVizError
from typing import List, Tuple, Any, Union, Dict
import importlib
import pkgutil
import shutil
from importlib_resources import files
import os
from pathlib import Path

plugin_prefix = "kmviz_"

class KmVizPlugin:
    def __init__(self):
        self._config: Dict[str, Any] = {}

    def providers(self) -> List[Tuple[str, Provider]]:
        """
        :returns: The providers implemented by the plugin, as list of tuples <name,'Provider'>
        """
        return []

    def databases(self) -> List[Tuple[str, MetaDB]]:
        """
        :returns: The metadbs implemented by the plugin, as list of tuples <name,'MetaDB'>
        """
        return []

    def notifiers(self) -> List[Tuple[str, Notifier]]:
        """
        :returns: The notifiers implemented by the plugin, as list of tuples <name, 'Notifier'>
        """

    def layouts(self) -> List[Tuple[str, Any, str]]:
        """
        :returns: The layouts implemented by the plugin, as list of tuples <name, dash_component, icon_name>
        """
        return []

    def external_scripts(self) -> List[Union[dict, str]]:
        """
        :returns: A list of js scripts to load, see https://dash.plotly.com/external-resources
        """
        return []

    def external_styles(self) -> List[Union[dict, str]]:
        """
        :returns: A list of css stylesheets to load, see https://dash.plotly.com/external-resources
        """
        return []

    def help(self) -> Any:
        """
        :returns: A Dash Component which will be displayed in the help tab.
        """
        return None

    def is_instance_plugin(self) -> bool:
        """
        :returns: True if the plugin is an instance plugin, False otherwise
        """
        return False

    def instance(self) -> Any:
        """
        :returns: A Dash Component which will be used as a homepage
        """
        return None

    def name(self) -> str:
        """
        :returns: The plugin name
        """
        return None

    def configure(self, config: Dict[str, Any]) -> None:
        self._config = config

    def has_api(self) -> str:
        """
        :returns: True if the plugin add API routes
        """
        return False

    def set_api(self, app):
        """
        Set API routes
        :returns: None
        """
        return None

    @property
    def config(self) -> Dict[str, Any]:
        return self._config


def installed_plugins() -> List[str]:
    res = []
    for finder, name, ispkg in pkgutil.iter_modules():
        if name.startswith("kmviz_"):
            res.append(name)
    return res

def search_for_plugin(name: str):
    module = importlib.import_module(name)
    p = module.kmviz_plugin

    if not isinstance(p, KmVizPlugin):
        raise KmVizError(f"Found module {name}, but 'kmviz_plugin' is not an instance of KmVizPlugin")
    return p

def copy_custom_assets(paths: List[str]):
    main = str(files("kmviz").joinpath("assets"))
    custom = f"{main}/_custom"

    if not paths:
        shutil.rmtree(custom, ignore_errors=True)
        return False

    os.makedirs(custom, exist_ok=True)

    for path in paths:
        path = Path(path)
        if path.exists():
            shutil.copy(path, custom)
        else:
            raise KmVizError(f"'{path}' not found.")
    return True


def copy_plugin_assets(plugin_name):
    p = files(plugin_name).joinpath("assets")
    main = str(files("kmviz").joinpath("assets"))
    if p.exists() and p.is_dir():
        shutil.copytree(p, f"{main}/_{plugin_name}_assets", dirs_exist_ok=True)
        return p
    return None

def search_for_plugins(modules: List[str]):
    plugins = {}

    for name in modules:
        if name.startswith("kmviz_"):
            try:
                module = importlib.import_module(name)
                plugins[name] = module.kmviz_plugin

                if not isinstance(plugins[name], KmVizPlugin):
                    raise KmVizError(f"Found module {name}, but 'kmviz_plugin' is not an instance of KmVizPlugin")
            except ImportError as e:
                raise KmVizError(f"Module '{name}' not found, installed plugins are [{','.join(installed_plugins())}]")

    return plugins


